#!/usr/bin/env python
"""Test calculation of CY14 static methods."""

from numpy.testing import assert_allclose

from pygmm import ChiouYoungs2014 as CY14


def test_depth_1_0():
    # Value calculated from NGAW2 spreadsheet
    assert_allclose(CY14.calc_depth_1_0(600, "california"), 0.1259, 4)
    assert_allclose(CY14.calc_depth_1_0(600, "japan"), 0.0331, 4)


def test_depth_tor():
    # Value calculated from NGAW2 spreadsheet
    assert_allclose(CY14.calc_depth_tor(6, "SS"), 2.2587685)
    assert_allclose(CY14.calc_depth_tor(6, "RS"), 6.3447262)



#!/usr/bin/env python
import json
import os

import numpy as np
import pytest
from numpy.testing import assert_allclose

from pygmm.baker_jayaram_2008 import calc_cond_mean_spectrum, calc_correls

fpath = os.path.join(os.path.dirname(__file__), "data", "baker_jayaram_2008.json")

with open(fpath) as fp:
    data = json.load(fp)


def idfn(case):
    if isinstance(case, dict):
        return "T_2={period_cond}s".format(**case)
    else:
        return case


@pytest.mark.parametrize("case", data["cases_correl"], ids=idfn)
def test_calc_correl(case):
    expected = case["correls"]
    actual = calc_correls(case["periods"], case["period_cond"]).tolist()
    # Points were digitized by hand so need to use larger tolerances
    assert_allclose(actual, expected, atol=0.005, rtol=0.005)


@pytest.mark.parametrize("case", data["cases_cms"], ids=idfn)
@pytest.mark.parametrize("param", ["psas_cms", "ln_stds_cms"])
def test_calc_cond_spectrum(case, param):
    model = data["model"]
    results = calc_cond_mean_spectrum(
        model["periods"],
        np.log(model["psas"]),
        model["ln_stds"],
        case["period_cond"],
        np.log(case["psa_cond"]),
    )
    # Need to go from ln_psa to psa
    actual = np.exp(results[0]) if param == "psas_cms" else results[1]
    expected = case[param]
    assert_allclose(actual, expected, atol=0.0001, rtol=0.0005)



import numpy as np
import pytest

from pygmm import AbrahamsonGregorAddo2016 as AGA16
from pygmm.model import Scenario

from . import load_tests

# Relative tolerance for all tests
RTOL = 2e-2

TESTS = load_tests("abrahamson_gregor_addo_2016.json.gz")


def create_model(params):
    s = Scenario(**params)
    m = AGA16(s)
    return m


@pytest.mark.parametrize("test", TESTS)
@pytest.mark.parametrize("key", ["spec_accels", "ln_stds", "pga", "ln_std_pga"])
def test_spec_accels(test, key):
    m = create_model(test["params"])
    np.testing.assert_allclose(getattr(m, key), test["results"][key], rtol=RTOL, err_msg=key)



import numpy as np
import pytest

from pygmm import CoppersmithBommer2014 as Model
from pygmm.model import Scenario

from . import load_tests

# Relative tolerance for all tests
RTOL = 2e-2

TESTS = load_tests("coppersmith_bommer_2014.json.gz")


def create_model(params):
    s = Scenario(**params)
    m = Model(s)
    return m


@pytest.mark.parametrize("test", TESTS)
@pytest.mark.parametrize("key", ["spec_accels", "ln_stds"])
def test_spec_accels(test, key):
    m = create_model(test["params"])
    np.testing.assert_allclose(getattr(m, key), test["results"][key], rtol=RTOL, err_msg=key)



import gzip
import json
import pathlib

FPATH_DATA = pathlib.Path(__file__).parent / "data"


def load_tests(fname):
    fpath = FPATH_DATA / fname
    if fpath.suffix == ".gz":
        fp = gzip.open(fpath, "rt")
    else:
        fp = fpath.open()

    return json.load(fp)



"""Bayless and Abrahamson (2019, :cite:`bayless19`) model."""

from typing import Optional, Union

import numpy as np

from . import model

# Based on code from Artie Rodgers
__author__ = "Albert Kottke"


class BaylessAbrahamson2019(model.Model):
    """Bayless and Abrahamson (2019, :cite:`bayless19`) model.

    This Fourier-amplitude spectra model was developed for active tectonic regions.

    Parameters
    ----------
    scenario : :class:`pygmm.model.Scenario`
        earthquake scenario

    """

    NAME = "Bayless and Abrahamson (2019)"
    ABBREV = "BA19"

    # Reference velocity (m/s)
    V_REF = 760.0

    # Load the coefficients for the model
    COEFF = model.load_data_file("bayless_abrahamson_2019.csv", 2)

    FREQS = COEFF["freq_hz"]

    IDX_5Hz = 170
    IDX_MAX = 238

    PARAMS = [
        model.NumericParameter("dist_rup", True, 0, 300),
        model.NumericParameter("mag", True, 3.0, 8.0),
        model.NumericParameter("v_s30", True, 180, 1500),
        model.NumericParameter("depth_tor", True, 0, 20),
        model.NumericParameter("depth_1_0", False),
        model.CategoricalParameter("mechanism", False, ["SS", "NS", "RS"], "SS"),
    ]

    def __init__(self, scenario: model.Scenario):
        """Initialize the model."""
        super().__init__(scenario)

        # Taek the reference intensity at 5 Hz
        ln_eas_ref = self._calc_ln_eas(None)
        self._ln_eas = self._calc_ln_eas(ln_eas_ref)
        self._ln_std = self._calc_ln_std()

    @property
    def freqs(self):
        return self.FREQS

    @property
    def ln_eas(self):
        return self._ln_eas

    @property
    def eas(self):
        return np.exp(self._ln_eas)

    @property
    def ln_std(self):
        return self._ln_std

    def _check_inputs(self) -> None:
        super()._check_inputs()
        s = self._scenario

        if s["depth_1_0"] is None:
            s["depth_1_0"] = self.calc_depth_1_0(s["v_s30"])

    def _calc_ln_eas(self, ln_eas_ref: Optional[float]) -> Union[float, np.ndarray]:
        """Compute the effective amplitude."""
        s = self._scenario

        if ln_eas_ref is None:
            # Compute the motion at the reference condition
            v_s30 = self.V_REF
            depth_1_0 = self.calc_depth_1_0(self.V_REF)
            # Only perform the 5 Hz calculation
            c = self.COEFF[self.IDX_5Hz]
        else:
            # Use site-specific properties
            v_s30 = s.v_s30
            depth_1_0 = s.depth_1_0
            # Only use the coefficients only the usable frequency range
            c = self.COEFF[: (self.IDX_MAX + 1)]

        f_M = (
            c.c1
            + c.c2 * (s.mag - 6)
            + ((c.c2 - c.c3) / c.cn) * np.log(1 + np.exp(c.cn * (c.cM - s.mag)))
        )

        f_P = (
            c.c4 * np.log(s.dist_rup + c.c5 * np.cosh(c.c6 * np.maximum(s.mag - c.chm, 0)))
            + (-0.5 - c.c4) * np.log(np.sqrt(s.dist_rup**2 + 50**2))
            + c.c7 * s.dist_rup
        )

        f_Ztor = c.c9 * min(s.depth_tor, 20)

        f_NM = c.c10 if s.mechanism == "NS" else 0

        c11 = np.select(
            [v_s30 <= 200, v_s30 <= 300, v_s30 <= 500], [c.c11a, c.c11b, c.c11c], c.c11d
        )
        f_Z1 = c11 * np.log((min(depth_1_0, 2) + 0.01) / (self.calc_depth_1_0(v_s30) + 0.01))

        ln_eas = f_M + f_P + f_Ztor + f_NM + f_Z1

        if ln_eas_ref is not None:
            # Add the site term
            ln_eas += self._calc_site_response(ln_eas_ref)

            # Extrapolate to 100 Hz using the kappa
            kappa = np.exp(-0.4 * np.log(v_s30 / 760) - 3.5)
            # Diminuition operator
            freq_max = self.FREQS[self.IDX_MAX]
            dimin = np.exp(-np.pi * kappa * (self.FREQS[(self.IDX_MAX + 1) :] - freq_max))
            ln_eas = np.r_[ln_eas, ln_eas[self.IDX_MAX] + np.log(dimin)]

        return ln_eas

    def _calc_site_response(self, ln_eas_ref: float) -> np.ndarray:
        v_s30 = self.scenario.v_s30
        # Only use the coefficients only the usable frequency range
        c = self.COEFF[: (self.IDX_MAX + 1)]

        I_R = np.exp(1.238 + 0.846 * ln_eas_ref)

        f_sl = c.c8 * np.log(min(v_s30, 1000) / 1000)

        f2 = c.f4 * (
            np.exp(c.f5 * (min(v_s30, self.V_REF) - 360)) - np.exp(c.f5 * (self.V_REF - 360))
        )

        f_nl = f2 * np.log((I_R + c.f3) / c.f3)

        # Use the minimum value at higher frequencies as described in the text and
        # referenced to Bayless and Abrahamson (2018b)
        i = np.argmin(f_nl)
        f_nl[(i + 1) :] = f_nl[i]
        width = 7
        kernel = np.ones(width) / width
        f_nl = np.convolve(f_nl, kernel, mode="same")
        # End needs to be forced down
        f_nl[-width:] = f_nl[i]

        f_s = f_sl + f_nl

        return f_s

    def _calc_ln_std(self) -> np.ndarray:
        """Compute the total standard deviation."""
        s = self._scenario
        c = self.COEFF

        def interp(left, right):
            return np.clip(left + (right - left) / 2 * (s.mag - 4), left, right)

        tau = interp(c.s1, c.s2)
        phi_s2s = interp(c.s3, c.s4)
        phi_ss = interp(c.s5, c.s6)

        ln_std = np.sqrt(tau**2 + phi_s2s**2 + phi_ss**2 + c.c1a**2)
        return ln_std

    @staticmethod
    def calc_depth_1_0(v_s30: float) -> float:
        """Compute the generic depth to 1 km/sec based on Vs30.

        Parameters
        ---------
        v_s30 : float
            site condition. Set `v_s30` to the reference
            velocity (e.g., 1180 m/s) for the reference response.

        Returns
        -------
        depth_1_0 : float
            depth to a shear-wave velocity of 1,000 m/sec
            (:math:`Z_{1.0}`, km).
        """
        power = 4
        v_ref = 610
        slope = -7.67 / power
        return (
            np.exp(slope * np.log((v_s30**power + v_ref**power) / (1360.0**power + v_ref**power)))
            / 1000
        )



"""Idriss (2014, :cite:`idriss14`) model."""

import numpy as np

from . import model

__author__ = "Albert Kottke"


class Idriss2014(model.GroundMotionModel):
    """Idriss (2014, :cite:`idriss14`) model.

    This model was developed for active tectonic regions as part of the
    NGA-West2 effort.

    Parameters
    ----------
    scenario : :class:`pygmm.model.Scenario`
        earthquake scenario

    """

    NAME = "Idriss (2014)"
    ABBREV = "I14"

    # Reference velocity (m/s)
    V_REF = 1200.0

    # Load the coefficients for the model
    COEFF = dict(
        small=model.load_data_file("idriss_2014-small.csv", 2),
        large=model.load_data_file("idriss_2014-large.csv", 2),
    )
    PERIODS = COEFF["small"]["period"]

    INDEX_PGA = 0
    INDICES_PSA = np.arange(22)

    PARAMS = [
        model.NumericParameter("dist_rup", True, None, 150),
        model.NumericParameter("mag", True, 5, None),
        model.NumericParameter("v_s30", True, 450, 1200),
        model.CategoricalParameter("mechanism", True, ["SS", "RS"], "SS"),
    ]

    def __init__(self, scenario: model.Scenario):
        """Initialize the model."""
        super().__init__(scenario)
        self._ln_resp = self._calc_ln_resp()
        self._ln_std = self._calc_ln_std()

    def _calc_ln_resp(self) -> np.ndarray:
        """Calculate the natural logarithm of the response.

        Returns
        -------
        ln_resp : class:`np.array`:
            natural log of the response

        """
        s = self._scenario
        c = self.COEFF["small"] if s.mag <= 6.75 else self.COEFF["large"]

        if s.mechanism == "RS":
            flag_mech = 1
        else:
            # SS/RS/U
            flag_mech = 0

        f_mag = c.alpha_1 + c.alpha_2 * s.mag + c.alpha_3 * (8.5 - s.mag) ** 2
        f_dst = -(c.beta_1 + c.beta_2 * s.mag) * np.log(s.dist_rup + 10) + c.gamma * s.dist_rup
        f_ste = c.epsilon * np.log(s.v_s30)
        f_mec = c.phi * flag_mech

        ln_resp = f_mag + f_dst + f_ste + f_mec

        return ln_resp

    def _calc_ln_std(self) -> np.ndarray:
        """Calculate the logarithmic standard deviation.

        Returns
        -------
        ln_std : class:`np.array`:
            natural log standard deviation

        """
        s = self._scenario
        ln_std = (
            1.18
            + 0.035 * np.log(np.clip(self.PERIODS, 0.05, 3.0))
            - 0.06 * np.clip(s.mag, 5.0, 7.5)
        )
        return ln_std



from pygmm import Scenario
from pygmm.afshari_stewart_2016 import AfshariStewart2016


def test_run():
    """Simple test to make sure we can get outputs."""

    s = Scenario(mag=6, dist_rup=50, v_s30=300, mechanism="SS")

    m = AfshariStewart2016(s)

    m.duration

    m.std_err



"""Pezeshk, Zandieh, and Tavakoli (2011, :cite:`pezeshk11`) model."""

import numpy as np

from . import model

__author__ = "Albert Kottke"


class PezeshkZandiehTavakoli2011(model.GroundMotionModel):
    """Pezeshk, Zandieh, and Tavakoli (2011, :cite:`pezeshk11`) model.

    Developed for the Eastern North America with a reference velocity of 2000
    m/s.

    Parameters
    ----------
    scenario : :class:`pygmm.model.Scenario`
        earthquake scenario

    """

    NAME = "Pezeshk et al. (2011)"
    ABBREV = "Pea11"

    # Reference shear-wave velocity (m/sec)
    V_REF = 2000.0

    # Load the coefficients for the model
    COEFF = model.load_data_file("pezeshk_zandieh_tavakoli_2011.csv", 1)
    PERIODS = COEFF["period"]

    INDEX_PGA = 0
    INDICES_PSA = np.arange(1, 23)

    PARAMS = [
        model.NumericParameter("mag", True, 5, 8),
        model.NumericParameter("dist_rup", True, None, 1000),
    ]

    def __init__(self, scenario: model.Scenario):
        """Initialize the model."""
        super().__init__(scenario)
        self._ln_resp = self._calc_ln_resp()
        self._ln_std = self._calc_ln_std()

    def _calc_ln_resp(self) -> np.ndarray:
        """Calculate the natural logarithm of the response.

        Returns
        -------
        ln_resp : class:`np.array`:
            natural log of the response

        """
        s = self._scenario
        c = self.COEFF

        dist = np.sqrt(s.dist_rup**2 + c.c_11**2)

        log10_resp = (
            c.c_1
            + c.c_2 * s.mag
            + c.c_3 * s.mag**2
            + (c.c_4 + c.c_5 * s.mag) * np.minimum(np.log10(dist), np.log10(70.0))
            + (c.c_6 + c.c_7 * s.mag)
            * np.maximum(np.minimum(np.log10(dist / 70.0), np.log10(140.0 / 70.0)), 0.0)
            + (c.c_8 + c.c_9 * s.mag) * np.maximum(np.log10(dist / 140.0), 0)
            + c.c_10 * dist
        )

        ln_resp = np.log(np.power(10, log10_resp))
        return ln_resp

    def _calc_ln_std(self) -> np.ndarray:
        """Calculate the logarithmic standard deviation.

        Returns
        -------
        ln_std : class:`np.array`:
            natural log standard deviation

        """
        s = self._scenario
        c = self.COEFF

        if s.mag <= 7.0:
            ln_std_mean = c.c_12 * s.mag + c.c_13
        else:
            ln_std_mean = -6.95e-3 * s.mag + c.c_14

        ln_std = np.sqrt(ln_std_mean**2 + c["sigma_reg"] ** 2)

        return ln_std



#!/usr/bin/env python
"""Test model interface using C03 model."""

import pytest

from pygmm.model import CategoricalParameter, NumericParameter, Parameter


def test_parameter():
    p = Parameter("test")
    p.check(None)


@pytest.fixture
def numeric_parameter():
    return NumericParameter("test", required=True, min_=0, max_=10)


@pytest.fixture
def categorical_parameter():
    return CategoricalParameter("test", required=True, options=["spam", "eggs"])


# See https://github.com/pytest-dev/pytest/issues/349
@pytest.fixture(params=["numeric_parameter", "categorical_parameter"])
def param(request):
    return request.getfixturevalue(request.param)


def test_required(param):
    with pytest.raises(ValueError):
        param.check(None)



#!/usr/bin/python
import gzip
import itertools
import json
import os

import numpy as np
import pytest

import pygmm

models = [
    pygmm.AbrahamsonSilvaKamai2014,
    pygmm.BooreStewartSeyhanAtkinson2014,
    pygmm.CampbellBozorgnia2014,
    pygmm.ChiouYoungs2014,
    pygmm.Idriss2014,
]

# Relative tolerance for all tests
RTOL = 1e-2

# Load the tests
fname = os.path.join(os.path.dirname(__file__), "data", "ngaw2_tests.json.gz")
with gzip.open(fname, "rt") as fp:
    tests = json.load(fp)

testdata = [(m, t["params"], t["results"][m.ABBREV]) for m, t in itertools.product(models, tests)]


@pytest.mark.parametrize("model,params,expected", testdata)
def test_ln_stds(model, params, expected):
    m = model(pygmm.model.Scenario(**params))
    np.testing.assert_allclose(
        m.interp_ln_stds(expected["periods"]),
        expected["ln_stds"],
        rtol=RTOL,
        err_msg="Logarithmic standard deviations",
    )


@pytest.mark.parametrize("model,params,expected", testdata)
def test_spec_accels(model, params, expected):
    m = model(pygmm.model.Scenario(**params))
    np.testing.assert_allclose(
        m.interp_spec_accels(expected["periods"]),
        expected["spec_accels"],
        rtol=RTOL,
        err_msg="Spectral accelerations",
    )


@pytest.mark.parametrize("model,params,expected", testdata)
@pytest.mark.parametrize("key", ["pga", "ln_std_pga", "pgv", "ln_std_pgv"])
def test_im_values(model, params, expected, key):
    if expected.get(key, None) is None:
        return

    m = model(pygmm.model.Scenario(**params))

    try:
        value = getattr(m, key)
    except NotImplementedError:
        return

    if value is None:
        return

    np.testing.assert_allclose(
        getattr(m, key),
        expected[key],
        rtol=RTOL,
    )



"""Abrahamson and Silva (1996, :cite:`abrahamson96`) duration model."""

import numpy as np

from . import model

__author__ = "Albert Kottke"


class AbrahamsonSilva1996(model.Model):
    """Abrahamson and Silva (1996, :cite:`abrahamson96`) duration model.

    Parameters
    ----------
    scenario : :class:`pygmm.model.Scenario`
        earthquake scenario

    """

    NAME = "Abrahamson Silva (1996)"
    ABBREV = "AS96"

    PARAMS = [
        model.NumericParameter("mag", True, 4, 7.5),
        model.NumericParameter("dist_rup", True, 0, 250),
        # FIXME add site_cond to Scenario
        model.CategoricalParameter("site_cond", True, ["soil", "rock"]),
    ]

    def __init__(self, scenario):
        super().__init__(scenario)

        s = self._scenario

        stress_drop = np.exp(5.204 + 0.851 * (s.mag - 6))
        moment = 10 ** (1.5 * s.mag + 16.05)

        self._ln_dur = np.log(
            (stress_drop / moment) ** (-1 / 3) / (4.9e6 * 3.2)
            + 0.805 * (1 if s.site_cond == "soil" else 0)
            + 0.063 * max(s.dist_rup - 10, 0)
        ) + self.calc_ln_dur_incr(0.75)
        self._std_err = 0.55

    @property
    def duration(self):
        return np.exp(self._ln_dur)

    @property
    def std_err(self):
        return self._std_err

    @staticmethod
    def calc_ln_dur_incr(nias):
        # Mask out inappropriate values
        nias = np.asarray(nias)
        mask = (nias < 0.10) | (0.95 < nias)
        nias[mask] = np.nan

        # Compute the increment due to the difference in the normalized Arias
        # intensity
        ln_i_ratio = np.log((nias - 0.05) / (1 - nias))
        ln_dur_incr = -0.532 + 0.552 * ln_i_ratio - 0.0262 * ln_i_ratio**2

        return ln_dur_incr

    def interp(self, nias, stds=None):
        # Mask out inappropriate values
        nias = np.asarray(nias)
        mask = (nias < 0.10) | (0.95 < nias)
        nias[mask] = np.nan

        ln_dur = self._ln_dur + self.calc_ln_dur_incr(nias)

        if stds is not None:
            std_errs = np.interp(
                nias,
                [
                    0.10,
                    0.15,
                    0.20,
                    0.25,
                    0.30,
                    0.35,
                    0.40,
                    0.45,
                    0.50,
                    0.55,
                    0.60,
                    0.65,
                    0.70,
                    0.75,
                    0.80,
                    0.85,
                    0.90,
                    0.95,
                ],
                [
                    0.843,
                    0.759,
                    0.713,
                    0.691,
                    0.674,
                    0.660,
                    0.646,
                    0.636,
                    0.628,
                    0.616,
                    0.605,
                    0.594,
                    0.582,
                    0.565,
                    0.545,
                    0.528,
                    0.510,
                    0.493,
                ],
            )
            ln_dur = ln_dur + np.array(stds)[:, np.newaxis] * std_errs

        return np.exp(ln_dur)



#!/usr/bin/env python
"""Test calculation of CB14 static methods."""

from numpy.testing import assert_allclose

from pygmm import CampbellBozorgnia2014 as CB14


def test_depth_2_5():
    # Value calculated from NGAW2 spreadsheet
    assert_allclose(CB14.calc_depth_2_5(600, "japan", None), 0.1844427)
    assert_allclose(CB14.calc_depth_2_5(600, "california", None), 0.7952589)



from pygmm import Scenario
from pygmm.kempton_stewart_2006 import KemptonStewart2006


def test_run():
    """Simple test to make sure we can get outputs."""

    s = Scenario(
        mag=6,
        dist_rup=50,
        v_s30=300,
    )

    m = KemptonStewart2006(s)

    m.duration

    m.std_err



"""Abrahamson, Silva, and Kamai (2014, :cite:`abrahamson14`) model."""

import numpy as np
from scipy.interpolate import interp1d

from . import model
from .types import ArrayLike

__author__ = "Albert Kottke"


class AbrahamsonSilvaKamai2014(model.GroundMotionModel):
    """Abrahamson, Silva, and Kamai (2014, :cite:`abrahamson14`) model.

    This model was developed for active tectonic regions as part of the
    NGA-West2 effort.

    Parameters
    ----------
    scenario : :class:`pygmm.model.Scenario`
        earthquake scenario

    """

    NAME = "Abrahamson, Silva, & Kamai (2014)"
    ABBREV = "ASK14"

    # Reference velocity (m/sec)
    V_REF = 1180.0

    # Load the coefficients for the model
    COEFF = model.load_data_file("abrahamson_silva_kamai_2014.csv", 2)

    PERIODS = COEFF["period"]

    INDICES_PSA = np.arange(22)
    INDEX_PGA = -2
    INDEX_PGV = -1

    PARAMS = [
        model.NumericParameter("dist_rup", True, None, 300),
        model.NumericParameter("dist_jb", True),
        model.NumericParameter("mag", True, 3, 8.5),
        model.NumericParameter("v_s30", True, 180, 1000),
        model.NumericParameter("depth_1_0", False),
        model.NumericParameter("depth_tor", False),
        model.NumericParameter("dip", True),
        model.NumericParameter("dist_crjb", False, default=15),
        model.NumericParameter("dist_x", False),
        model.NumericParameter("dist_y0", False),
        model.NumericParameter("width", False),
        model.CategoricalParameter("mechanism", True, ["SS", "NS", "RS"]),
        model.CategoricalParameter(
            "region",
            False,
            ["global", "california", "china", "italy", "japan", "taiwan"],
            "global",
        ),
        model.CategoricalParameter("vs_source", False, ["measured", "inferred"], "measured"),
        model.CategoricalParameter("is_aftershock", False, [True, False], False),
        model.CategoricalParameter("on_hanging_wall", False, [True, False], False),
    ]

    def _check_inputs(self) -> None:
        """Check the inputs."""
        super()._check_inputs()
        s = self._scenario
        if s["width"] is None:
            s["width"] = self.calc_width(s.mag, s.dip)

        if s["depth_tor"] is None:
            s["depth_tor"] = self.calc_depth_tor(s.mag)

    def __init__(self, scenario: model.Scenario):
        """Initialize the model."""
        super().__init__(scenario)
        # Compute the response at the reference velocity
        resp_ref = np.exp(self._calc_ln_resp(self.V_REF, np.nan))

        self._ln_resp = self._calc_ln_resp(self._scenario.v_s30, resp_ref)
        self._ln_std, self._tau, self._phi = self._calc_ln_std(resp_ref)

    def _calc_ln_resp(self, v_s30: float, resp_ref: ArrayLike) -> np.ndarray:
        """Calculate the natural logarithm of the response.

        Parameters
        ----------
        v_s30 : float
            site condition. Set `v_s30` to the reference
            velocity (e.g., 1180 m/s) for the reference response.
        resp_ref :  array_like, optional
            response at the reference condition. Required if `v_s30` is not
            equal to reference velocity.

        Returns
        -------
        ln_resp: :class:`np.ndarray`
            natural log of the response.

        """
        c = self.COEFF
        s = self._scenario

        # Magnitude scaling
        f1 = self._calc_f1()

        if s.on_hanging_wall:
            # Hanging-wall term
            f4 = self._calc_f4()
        else:
            f4 = 0

        # Depth to top of rupture term
        f6 = c.a15 * np.clip(s.depth_tor / 20, 0, 1)
        # Style of faulting
        if s.mechanism == "RS":
            f7 = c.a11 * np.clip(s.mag - 4, 0, 1)
            f8 = 0
        elif s.mechanism == "NS":
            f7 = 0
            f8 = c.a12 * np.clip(s.mag - 4, 0, 1)
        else:
            f7, f8 = 0, 0

        site_term = self.calc_site_term(resp_ref, v_s30, s.depth_1_0, s.region)

        # Aftershock term
        if s.is_aftershock:
            f11 = c.a14 * np.clip(1 - (s.dist_crjb - 5) / 10, 0, 1)
        else:
            f11 = 0

        if s.region == "taiwan":
            v_1 = np.exp(-0.35 * np.log(np.clip(c.period, 0.5, 3) / 0.5) + np.log(1500))
            vs_ratio = np.minimum(v_s30, v_1) / c.v_lin
            freg = c.a31 * np.log(vs_ratio) + c.a25 * s.dist_rup
        elif s.region == "china":
            freg = c.a28 * s.dist_rup
        elif s.region == "japan":
            f13 = interp1d(
                [150, 250, 350, 450, 600, 850, 1150],
                np.c_[c.a36, c.a37, c.a38, c.a39, c.a40, c.a41, c.a42],
                copy=False,
                bounds_error=False,
                fill_value=(c.a36, c.a42),
            )(v_s30)
            freg = f13 + c.a29 * s.dist_rup
        else:
            freg = 0

        return f1 + f4 + f6 + f7 + f8 + f11 + freg + site_term

    @classmethod
    def calc_site_term(
        cls,
        resp_ref: ArrayLike,
        v_s30: float,
        depth_1_0: float,
        region: str = "california",
    ):
        """Calculate the site term, which includes site and basin effects.

        Parameters
        ----------
        resp_ref :  array_like, optional
            response at the reference condition
        v_s30 : float
            site condition. Set `v_s30` to the reference
            velocity (e.g., 1180 m/s) for the reference response.
        depth_1_0 : float
            depth to the 1.0 km∕s shear-wave velocity horizon beneath the site,
            :math:`Z_{1.0}` in (km).
        region : str, optional
            region of basin model. Valid options: 'california', 'japan'. If
            *None*, then 'california' is used as the default value.

        Returns
        -------
        site_term: :class:`np.ndarray`
            site term that is applied to the natural log response.
        """
        c = cls.COEFF

        v_1 = np.exp(-0.35 * np.log(np.clip(c.period, 0.5, 3) / 0.5) + np.log(1500))

        vs_ratio = np.minimum(v_s30, v_1) / c.v_lin
        # Linear site model
        f5 = (c.a10 + c.b * c.n) * np.log(vs_ratio)
        # Nonlinear model
        mask = vs_ratio < 1
        f5[mask] = (
            c.a10 * np.log(vs_ratio)
            - c.b * np.log(resp_ref + c.c)
            + c.b * np.log(resp_ref + c.c * vs_ratio**c.n)
        )[mask]

        # Basin term
        if v_s30 == cls.V_REF or depth_1_0 is None:
            # No basin response
            f10 = 0
        else:
            # Ratio between site depth_1_0 and model center
            ln_depth_ratio = np.log((depth_1_0 + 0.01) / (cls.calc_depth_1_0(v_s30, region) + 0.01))
            slope = interp1d(
                [150, 250, 400, 700],
                np.c_[c.a43, c.a44, c.a45, c.a46],
                copy=False,
                bounds_error=False,
                fill_value=(c.a43, c.a46),
            )(v_s30)
            f10 = slope * ln_depth_ratio

        return f5 + f10

    def _calc_ln_std(self, psa_ref: ArrayLike) -> (np.ndarray, np.ndarray, np.ndarray):
        """Calculate the logarithmic standard deviation.

        Parameters
        ----------
        psa_ref : array_like
           spectral accelerations at the reference condition

        Returns
        -------
        ln_std :  :class:`np.ndarray`
            logarithmic standard deviation.
        """
        s = self._scenario
        c = self.COEFF

        if s.region == "japan":
            phi_al = c.s5 + (c.s6 - c.s5) * np.clip((s.dist_rup - 30) / 50, 0, 1)
        else:
            transition = np.clip((s.mag - 4) / 2, 0, 1)
            if s.vs_source == "measured":
                phi_al = c.s1m + (c.s2m - c.s1m) * transition
            else:
                phi_al = c.s1e + (c.s2e - c.s1e) * transition

        tau_al = c.s3 + (c.s4 - c.s3) * np.clip((s.mag - 5) / 2, 0, 1)
        tau_b = tau_al

        # Remove period independent site amplification uncertainty of 0.4
        phi_amp = 0.4
        phi_b = np.sqrt(np.maximum(phi_al**2 - phi_amp**2, 0))

        # The partial derivative of the amplification with respect to
        # the reference intensity
        deriv = (-c.b * psa_ref) / (psa_ref + c.c) + (c.b * psa_ref) / (
            psa_ref + c.c * (s.v_s30 / c.v_lin) ** c.n
        )
        deriv[s.v_s30 >= c.v_lin] = 0
        tau = tau_b * (1 + deriv)
        phi = np.sqrt(phi_b**2 * (1 + deriv) ** 2 + phi_amp**2)

        ln_std = np.sqrt(phi**2 + tau**2)
        return ln_std, tau, phi

    @staticmethod
    def calc_width(mag: float, dip: float) -> float:
        r"""Compute the fault width based on equation in NGW2 spreadsheet.

        This equation is not provided in the paper.

        Parameters
        ----------
        mag : float
            moment magnitude of the event (:math:`M_w`)
        dip : float
            Fault dip angle (:math:`\phi`, deg)

        Returns
        -------
        width : float
            estimated fault width (:math:`W`, km)
        """
        return min(18 / np.sin(np.radians(dip)), 10 ** (-1.75 + 0.45 * mag))

    @staticmethod
    def calc_depth_tor(mag: float) -> float:
        """Calculate the depth to top of rupture (km).

        Parameters
        ----------
        mag : float
            moment magnitude of the event (:math:`M_w`)

        Returns
        -------
        depth_tor : float
            estimated depth to top of rupture (km)
        """
        return np.interp(mag, [5.0, 7.2], [7.8, 0])

    @staticmethod
    def calc_depth_1_0(v_s30: float, region: str = "california") -> float:
        """Estimate the depth to 1 km/sec horizon (:math:`Z_{1.0}`) based on
        :math:`V_{s30}` and region.

        This is based on equations 18 and 19 in the :cite:`abrahamson14`
        and differs from the equations in the :cite:`chiou14`.

        Parameters
        ----------
        v_s30 : float
            time-averaged shear-wave velocity over the top 30 m
            of the site (:math:`V_{s30}`, m/s).
            Keyword Args:
        region : str, optional
            region of basin model. Valid options: 'california', 'japan'. If
            *None*, then 'california' is used as the default value.

        Returns
        -------
        depth_1_0 : float
            depth to a shear-wave velocity of 1,000 m/sec
            (:math:`Z_{1.0}`, km).

        """
        if region in ["japan"]:
            # Japan
            power = 2
            v_ref = 412
            slope = -5.23 / power
        else:
            # Global
            power = 4
            v_ref = 610
            slope = -7.67 / power

        return (
            np.exp(slope * np.log((v_s30**power + v_ref**power) / (1360.0**power + v_ref**power)))
            / 1000
        )

    def _calc_f1(self) -> np.ndarray:
        """Calculate the magnitude scaling parameter f1."""
        c = self.COEFF
        s = self._scenario

        # Magnitude dependent taper
        dist = np.sqrt(s.dist_rup**2 + (c.c4 - (c.c4 - 1) * np.clip(5 - s.mag, 0, 1)) ** 2)

        # Magnitude scaling
        # Need to copy c.a1 to that it isn't modified during the following
        # operations.
        f1 = np.array(c.a1)
        ma1 = s.mag <= c.m2
        f1[ma1] += (
            c.a4 * (c.m2 - c.m1)
            + c.a8 * (8.5 - c.m2) ** 2
            + c.a6 * (s.mag - c.m2)
            + c.a7 * (s.mag - c.m2) ** 2
            + (c.a2 + c.a3 * (c.m2 - c.m1)) * np.log(dist)
            + c.a17 * s.dist_rup
        )[ma1]

        f1[~ma1] += (
            c.a8 * (8.5 - s.mag) ** 2
            + (c.a2 + c.a3 * (s.mag - c.m1)) * np.log(dist)
            + c.a17 * s.dist_rup
        )[~ma1]

        ma2 = np.logical_and(~ma1, s.mag <= c.m1)
        f1[ma2] += (c.a4 * (s.mag - c.m1))[ma2]

        ma3 = np.logical_and(~ma1, s.mag > c.m1)
        f1[ma3] += (c.a5 * (s.mag - c.m1))[ma3]

        return f1

    def _calc_f4(self) -> np.ndarray:
        """Calculate the hanging-wall parameter f4."""
        # Move this into a decorator?
        c = self.COEFF
        s = self._scenario

        t1 = min(90 - s.dip, 60) / 45
        # Constant from page 1041
        a2hw = 0.2
        if s.mag <= 5.5:
            t2 = 0
        elif s.mag < 6.5:
            t2 = 1 + a2hw * (s.mag - 6.5) - (1 - a2hw) * (s.mag - 6.5) ** 2
        else:
            t2 = 1 + a2hw * (s.mag - 6.5)

        # Constants defined on page 1040
        r1 = s.width * np.cos(np.radians(s.dip))
        r2 = 3 * r1
        h1 = 0.25
        h2 = 1.5
        h3 = -0.75
        if s.dist_x < r1:
            t3 = h1 + h2 * (s.dist_x / r1) + h3 * (s.dist_x / r1) ** 2
        elif s.dist_x < r2:
            t3 = 1 - ((s.dist_x - r1) / (r2 - r1))
        else:
            t3 = 0

        t4 = np.clip(1 - s.depth_tor**2 / 100, 0, 1)

        if s.dist_y0 is None:
            t5 = np.clip(1 - s.dist_jb / 30, 0, 1)
        else:
            dist_y1 = s.dist_x * np.tan(np.radians(20))
            t5 = np.clip(1 - (s.dist_y0 - dist_y1) / 5, 0, 1)

        f4 = c.a13 * t1 * t2 * t3 * t4 * t5

        return f4



"""Derras, Bard and Cotton (2014, :cite:`derras14`) model."""

import json
import os

import numpy as np

from . import model

__author__ = "Albert Kottke"


class DerrasBardCotton2014(model.GroundMotionModel):
    """Derras, Bard and Cotton (2014, :cite:`derras14`) model.

    Parameters
    ----------
    scenario : :class:`pygmm.model.Scenario`
        earthquake scenario

    """

    NAME = "Derras, Bard & Cotton (2014)"
    ABBREV = "DBC13"

    # Load the coefficients for the model
    COEFF = json.load(
        open(os.path.join(os.path.dirname(__file__), "data", "derras_bard_cotton_2014.json"))
    )
    GRAVITY = 9.80665
    PERIODS = np.array(COEFF["period"])

    INDICES_PSA = np.arange(2, 64)
    INDEX_PGA = 1
    INDEX_PGV = 0
    PARAMS = [
        model.NumericParameter("dist_jb", True, 5, 200),
        model.NumericParameter("mag", True, 4, 7),
        model.NumericParameter("v_s30", True, 200, 800),
        model.NumericParameter("depth_hyp", True, 0, 25),
        model.CategoricalParameter("mechanism", True, ["SS", "NS", "RS"]),
    ]

    def __init__(self, scenario: model.Scenario):
        """Initialize the model."""
        super().__init__(scenario)
        c = self.COEFF
        # Values modified during the calculation
        s = dict(self._scenario)

        for k in ["v_s30", "dist_jb"]:
            s["log10_" + k] = np.log10(s[k])
        # Translate to mechanism integer
        s["mechanism"] = dict(NS=1, RS=3, SS=4)[s["mechanism"]]

        # Create the normalized parameter matrix
        keys = ["log10_dist_jb", "mag", "log10_v_s30", "depth_hyp", "mechanism"]
        values = np.array([s[k] for k in keys])
        limits = np.rec.array([c["min_max"][k] for k in keys], names="min,max")
        p_n = np.array(2 * (values - limits["min"]) / (limits["max"] - limits["min"]) - 1).T

        # Compute the normalized response
        b_1 = np.array(c["b_1"]).T
        b_2 = np.array(c["b_2"]).T
        w_1 = np.array(c["w_1"])
        w_2 = np.array(c["w_2"])

        log10_resp_n = b_2 + w_2 @ np.tanh(b_1 + w_1 @ p_n)

        # Convert from normalized values
        log10_resp_limits = np.rec.array(c["min_max"]["log10_resp"], names="min,max")
        log10_resp = (
            0.5 * (log10_resp_n + 1) * (log10_resp_limits["max"] - log10_resp_limits["min"])
            + log10_resp_limits["min"]
        )
        # Convert from m/sec and m/sec² into cm/sec and g
        scale = np.log10(np.r_[0.01, self.GRAVITY * np.ones(self.PERIODS.size - 1)])

        self._ln_resp = np.log(10 ** (log10_resp - scale))
        self._ln_std = np.log(10 ** np.array(c["log10_std"]["total"]))



"""Afshari and Stewart (2016, :cite:`afshari2016`) duration model."""

import numpy as np

from . import model


class AfshariStewart2016(model.Model):
    """Afshari and Stewart (2016, :cite:`afshari2016`) duration model.

    Parameters
    ----------
    scenario : :class:`pygmm.model.Scenario`
        earthquake scenario

    """

    NAME = "Afshari and Stewart (2016)"
    ABBREV = "AS16"

    PARAMS = [
        model.NumericParameter("mag", True, 3, 7.9),
        model.NumericParameter("dist_rup", True, 0, 200),
        model.NumericParameter("v_s30", True, 200, 1000),
        model.CategoricalParameter("mechanism", True, ["SS", "NS", "RS"]),
        model.NumericParameter("depth_1_0", False),
    ]

    def __init__(self, scenario):
        super().__init__(scenario)

        # scenario
        s = self._scenario

        # duration D_{5-75}, D_{5-95}, D_{20-80},

        # source coefficients
        mag_1 = np.array([5.35, 5.2, 5.2])
        mag_2 = np.array([7.15, 7.4, 7.4])
        mag_star = 6
        if s.mechanism is None:
            b_0 = np.array([1.280, 2.182, 0.8822])
            b_1 = np.array([5.576, 3.628, 6.182])
        elif s.mechanism == "NS":
            b_0 = np.array([1.555, 2.541, 1.409])
            b_1 = np.array([4.992, 3.170, 4.778])
        elif s.mechanism == "RS":
            b_0 = np.array([0.7806, 1.612, 0.7729])
            b_1 = np.array([7.061, 4.536, 6.188])
        elif s.mechanism == "SS":
            b_0 = np.array([1.279, 2.302, 0.8804])
            b_1 = np.array([5.578, 3.467, 6.188])
        b_2 = np.array([0.9011, 0.9443, 0.7414])
        b_3 = np.array([-1.684, -3.911, -3.164])
        # stress drop indices for duration measures

        stress_drop = np.exp(
            b_1 + b_2 * (np.minimum(s.mag, mag_2) - mag_star) + b_3 * np.maximum((s.mag - mag_2), 0)
        )
        # corner frequency
        f_0 = self._corner_freq(stress_drop)

        # source duration
        F_E = (s.mag <= mag_1) * b_0 + (s.mag > mag_1) * (1 / f_0)

        # path coeficients
        c_1 = np.array([0.1159, 0.3165, 0.0646])
        c_2 = np.array([0.1065, 0.2539, 0.0865])
        c_3 = np.array([0.0682, 0.0932, 0.0373])
        c_4 = np.array([-0.2246, -0.3183, -0.4237])
        c_5 = np.array([0.0006, 0.0006, 0.0005])
        R_1 = 10
        R_2 = 50
        # path duration
        F_P = (
            c_1 * np.minimum(s.dist_rup, R_1)
            + c_2 * np.maximum((np.minimum(s.dist_rup, R_2) - R_1), 0)
            + c_3 * np.maximum((s.dist_rup - R_2), 0)
        )

        # site coefficients
        V_1 = 600
        V_ref = np.array([368.2, 369.9, 369.6])
        dz_1ref = 200
        # depth to bedrock duration
        if s.depth_1_0 is not None:
            dz_1 = s.depth_1_0 - self.calc_depth_1_0(s.v_s30, s.mechanism)
            F_dz1 = c_5 * (dz_1 if dz_1 <= dz_1ref else dz_1ref)
        else:
            F_dz1 = 0
        # site duration
        F_S = c_4 * np.log(np.minimum(s.v_s30, V_1) / V_ref) + F_dz1

        # aleatory coefficients
        tau_1 = np.array([0.28, 0.25, 0.30])
        tau_2 = np.array([0.25, 0.19, 0.19])
        phi_1 = np.array([0.54, 0.43, 0.56])
        phi_2 = np.array([0.41, 0.35, 0.45])
        # aleatory variability
        tau = tau_1 + (tau_2 - tau_1) * (np.minimum(np.maximum(s.mag, 6.5), 7.0) - 6.5) / (
            7.0 - 6.5
        )
        phi = phi_1 + (phi_2 - phi_1) * (np.minimum(np.maximum(s.mag, 5.5), 5.75) - 5.5) / (
            5.75 - 5.50
        )

        # total druation
        self._ln_dur = np.log(F_E + F_P) + F_S

        # aleatory standard deviation
        self._std_err = np.sqrt(tau**2 + phi**2)

    @staticmethod
    def calc_depth_1_0(v_s30: float, region: str = "california") -> float:
        if region in ["japan"]:
            # Japan
            power = 2
            v_ref = 412.39
            slope = -5.23 / power
        else:
            # Global
            power = 4
            v_ref = 570.94
            slope = -7.15 / power

        return (
            np.exp(slope * np.log((v_s30**power + v_ref**power) / (1360.0**power + v_ref**power)))
            / 1000
        )

    @property
    def duration(self):
        """Return the durations as a `np.recarray`. Values can be accessed with
        'D_5t75', 'D_5t95', or 'D_20t80'."""
        return AfshariStewart2016._as_recarray(np.exp(self._ln_dur))

    @property
    def std_err(self):
        """Return the standard errors of the durations as a `np.recarray`. Values can be
        accessed with 'D_5t75', 'D_5t95', or 'D_20t80'."""
        return AfshariStewart2016._as_recarray(self._std_err[:3])

    @staticmethod
    def _as_recarray(values):
        return np.rec.fromarrays(
            values,
            names=[
                "D_5t75",
                "D_5t95",
                "D_20t80",
            ],
        )

    def _corner_freq(self, stress_drop):
        """Corner frequency."""
        moment = 10 ** (1.5 * self.scenario.mag + 16.05)
        return (4.9e6 * 3.2) * (stress_drop / moment) ** (1 / 3)



"""Gülerce and Abrahamson (2011, :cite:`gulerce11`) V/H model."""

import numpy as np

from . import model
from .abrahamson_silva_kamai_2014 import AbrahamsonSilvaKamai2014 as ASK14
from .types import ArrayLike

__author__ = "Albert Kottke"


class GulerceAbrahamson2011(model.Model):
    """Gülerce and Abrahamson (2011, :cite:`gulerce11`) V/H model.

    This model was developed for active tectonic regions.

    Parameters
    ----------
    scenario : :class:`pygmm.model.Scenario`
        earthquake scenario

    """

    NAME = "Gülerce & Abrahamson (2011)"
    ABBREV = "GA11"

    # Reference velocity (m/s)
    V_REF = 1100.0

    # Load the coefficients for the model
    COEFF = model.load_data_file("gulerce_abrahamson_2011-coefficients.csv", 2)
    PERIODS = COEFF["period"]

    INDEX_PGA = 0
    INDEX_PGV = 1
    INDICES_PSA = 2 + np.arange(22)

    PARAMS = [
        model.NumericParameter("dist_rup", True, None, 200),
        model.NumericParameter("mag", True, 5, 8.5),
        model.NumericParameter("v_s30", True, 450, 1200),
        model.NumericParameter("pga_ref", True, 0.001, 10),
        model.CategoricalParameter("mechanism", True, ["SS", "NS", "RS"]),
    ]

    def __init__(self, *args, **kwargs):
        """Initialize the model."""
        super().__init__(*args, **kwargs)

        if "pga_ref" not in self._scenario:
            s = model.Scenario(self._scenario)
            s["v_s30"] = self.V_REF
            m = ASK14(s)
            self._scenario["pga_ref"] = m.pga

        self._ln_ratio = self._calc_ln_ratio()
        self._ln_std = self._calc_ln_std()

    @property
    def ratio(self):
        return np.exp(self._ln_ratio)

    @property
    def ln_std(self):
        return self._ln_std

    def interp_ratio(self, periods: ArrayLike) -> np.ndarray:
        """Interpolate the spectral ratio.

        Interpolation of the spectral ratio done in log-period and log-ratio space.

        Parameters
        ----------
        periods : array_like
            spectral periods to interpolate the response.

        Returns
        -------
        ratio : np.ndarray
            interpolated ratio

        """

        return np.exp(
            np.interp(
                np.log(periods),
                np.log(self.PERIODS[self.INDICES_PSA]),
                self._ln_ratio[self.INDICES_PSA],
            )
        )

    def interp_ln_std(self, periods: ArrayLike) -> np.ndarray:
        """Interpolate the spectral ratio.

        Interpolation of the spectral ratio done in log-period space.

        Parameters
        ----------
        periods : array_like
            spectral periods to interpolate the response.

        Returns
        -------
        ratio : np.ndarray
            interpolated ratio

        """

        return np.exp(
            np.interp(
                np.log(periods),
                np.log(self.PERIODS[self.INDICES_PSA]),
                self._ln_std[self.INDICES_PSA],
            )
        )

    def _calc_ln_ratio(self):
        s = self._scenario
        c = self.COEFF

        dist = np.sqrt(s.dist_rup**2 + c.c4**2)

        # Magnitude and distance scaling
        f1 = (
            c.a1
            + c.a8 * (8.5 - s.mag) ** 2
            + (c.a2 + c.a3 * (s.mag - c.c1)) * np.log(dist)
            + np.select([(s.mag < c.c1), True], [c.a4 * (s.mag - c.c1), c.a5 * (s.mag - c.c1)])
        )
        # Site model
        v_1 = np.select(
            [
                self.PERIODS < 0,
                self.PERIODS <= 0.50,
                self.PERIODS <= 1.00,
                self.PERIODS <= 2.00,
                self.PERIODS <= 10.0,
            ],
            # Use minimum on period to prevent runtime warning.
            [
                862,
                1500.0,
                np.exp(8 - 0.795 * np.log(np.maximum(0.001, self.PERIODS) / 0.21)),
                np.exp(6.76 - 0.297 * np.log(np.maximum(0.001, self.PERIODS))),
                700,
            ],
        )
        v_s30_lim = np.minimum(s.v_s30, v_1)
        vs_ratio = v_s30_lim / c.v_lin
        f5 = c.a10 * np.log(vs_ratio) - np.select(
            [v_s30_lim < c.v_lin, True],
            [
                -c.b * np.log(s.pga_ref + c.c) + c.b * np.log(s.pga_ref + c.c * vs_ratio**c.n),
                c.b * c.n * np.log(vs_ratio),
            ],
        )

        ln_ratio = f1 + f5
        if s.mechanism == "RS":
            ln_ratio += c.a6
        elif s.mechanism == "NS":
            ln_ratio += c.a7

        return ln_ratio

    def _calc_ln_std_within(self):
        c = self.COEFF
        mag = self._scenario.mag
        return np.select(
            [mag < 5, mag <= 7, True],
            [c.s1, c.s1 + 0.5 * (c.s2 - c.s1) * (mag - 5), c.s2],
        )

    def _calc_ln_std_between(self):
        c = self.COEFF
        mag = self._scenario.mag
        return np.select(
            [mag < 5, mag <= 7, True],
            [c.s3, c.s3 + 0.5 * (c.s4 - c.s3) * (mag - 5), c.s4],
        )

    def _calc_ln_std(self):
        between = self._calc_ln_std_between()
        within = self._calc_ln_std_within()

        return np.sqrt(between**2 + within**2)



#!/usr/bin/env python
"""Test calculation of CY14 static methods."""

import csv

import numpy as np
import pytest
from numpy.testing import assert_allclose
from scipy.constants import g as GRAVITY

from pygmm import BaylessAbrahamson2019 as BA19

from . import FPATH_DATA

# Convert from cm/s to g-s
TO_G = 1 / (GRAVITY * 100)


def read_scenarios(fpath):
    with fpath.open() as fp:
        reader = csv.reader(fp)
        rows = list(reader)

    # Read the parameters
    stop = 6
    parser = {"mechanism": str}
    scenarios = [
        {r[0]: parser.get(r[0], float)(r[i + 1]) for r in rows[:stop]}
        for i in range(len(rows[0]) - 1)
    ]

    # Read the spectrum
    start = 8
    freqs = np.array([float(r[0]) for r in rows[start:]])
    for i, s in enumerate(scenarios):
        s["freqs"] = freqs
        s["eas_median"] = np.array([float(r[i + 1]) for r in rows[start:]])

    return scenarios


def iter_scenarios():
    for fpath in FPATH_DATA.glob("BA19-*.csv"):
        yield from read_scenarios(fpath)


@pytest.mark.parametrize(
    "scenario", iter_scenarios(), ids=lambda s: f"M {s['mag']}, Vs30 {s['v_s30']} m/s"
)
def test_ba19(scenario):
    m = BA19(scenario)
    assert_allclose(m.freqs, scenario["freqs"])
    assert_allclose(m.eas, scenario["eas_median"] * TO_G, rtol=0.005)



#!/usr/bin/env python
"""Create test cases for the NGA West2 models."""

import gzip
import json
import os

import xlwings as xw


def iter_parameters(parameters):
    max_len = max(len(p) for p in parameters.values())

    for i in range(max_len):
        yield {k: v[i % len(v)] for k, v in parameters.items()}


def reformat_params(params):
    """Transform from pygmm input to spreadsheet input."""
    if "mechanism" in params:
        # Replace mechanism with flags
        mech = params.pop("mechanism")
        if mech == "SS":
            params["flag_u"] = 0
            params["flag_rv"] = 0
            params["flag_nm"] = 0
        elif mech == "NS":
            params["flag_u"] = 0
            params["flag_rv"] = 0
            params["flag_nm"] = 1
        elif mech == "RS":
            params["flag_u"] = 0
            params["flag_rv"] = 1
            params["flag_nm"] = 0
        elif mech == "U":
            params["flag_u"] = 1
            params["flag_rv"] = 0
            params["flag_nm"] = 0

    try:
        on_hanging_wall = params.pop("on_hanging_wall")
        params["flag_hw"] = 1 if on_hanging_wall else 0
    except KeyError:
        pass

    try:
        params["flag_meas"] = params.pop("vs_source")
    except KeyError:
        pass

    try:
        params["region"] = params["region"].replace("_", " ").title()
    except KeyError:
        pass

    # Replace None with 999
    for key in ["depth_tor", "depth_hyp", "depth_1_0", "depth_2_5", "width", "dist_y0"]:
        if key in params and params[key] is None:
            params[key] = 999

    if "dist_rup" not in params and "dist_jb" in params:
        params["dist_rup"] = params["dist_jb"]

    try:
        # No supported to changing the DPP
        params.pop("dpp_centered")
    except KeyError:
        pass

    return params


def load_params(wb, **params):
    params = reformat_params(dict(params))
    for k, v in params.items():
        print("Setting:", k, CELLS[k], v)
        xw.Range("Main", CELLS[k], wkb=wb).value = v

    # Force a re-calculation. This is the same as Ctrl + Alt + F9
    wb.xl_app.CalculateFull()


def get_results(wb, abbrev):
    sheetname = abbrev
    result_columns = RESULT_COLUMNS[abbrev]
    # Collect response spectrum
    d = {}
    for key, rc in zip(["periods", "spec_accels", "ln_stds"], result_columns):
        d[key] = xw.Range(sheetname, "{col}6:{col}26".format(col=rc), wkb=wb).value

    # Collect the PGA and PGV
    for key, row in zip(["pga", "pgv"], [28, 29]):
        values = xw.Range(
            sheetname,
            "{col0:}{row:d}:{col1:}{row:d}".format(
                col0=result_columns[1], col1=result_columns[2], row=row
            ),
            wkb=wb,
        ).value
        for subkey, v in zip(["", "_ln_std"], values):
            try:
                float(v)
            except ValueError:
                v = None
            d[key + subkey] = v

    return d


wb_fname = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "NGAW2_GMPE_Spreadsheets_v5.7_041415_Protected.xlsm")
)

# Parameters
params_site = {
    "v_s30": [300, 450, 760, 1000],
    "depth_1_0": [0.4, 0.3, 0.04, 0.005],
    "depth_2_5": [1.8, 1.1, 0.4, 0.3],
    "vs_source": ["measured", "inferred"],
    "region": [
        "global",
        "california",
        "china",
        "japan",
        "italy",
        "new_zealand",
        "turkey",
        "taiwan",
    ],
}

params_fault = [
    {
        "mag": 6.0,
        "dist_rup": 3.16,
        "dist_jb": 1.0,
        "dist_x": 1.0,
        "dist_y0": None,
        "mechanism": "SS",
        "on_hanging_wall": False,
        "dip": 90,
        "depth_tor": 0,
        "depth_hyp": 8,
        "width": 10,
    },
    {
        "mag": 7.1,
        "dist_rup": 10,
        "dist_jb": 2,
        "dist_x": 16,
        "dist_y0": None,
        "mechanism": "RS",
        "on_hanging_wall": True,
        "dip": 50,
        "depth_tor": 1,
        "depth_hyp": 8,
        "width": 15,
    },
    {
        "mag": 4.5,
        "dist_rup": 10,
        "dist_jb": 8.0,
        "dist_x": -8,
        "dist_y0": None,
        "mechanism": "RS",
        "on_hanging_wall": False,
        "dip": 60,
        "depth_tor": 4,
        "depth_hyp": 5,
        "width": 2,
    },
    {
        "mag": 6.0,
        "dist_rup": 5,
        "dist_jb": 0.0,
        "dist_x": 0.1,
        "dist_y0": None,
        "mechanism": "NS",
        "on_hanging_wall": True,
        "dip": 60,
        "depth_tor": 4,
        "depth_hyp": 5,
        "width": 8,
    },
    {
        "mag": 6.5,
        "dist_rup": 100,
        "dist_jb": 100,
        "dist_x": 60,
        "dist_y0": 70,
        "mechanism": "U",
        "on_hanging_wall": False,
        "dip": 90,
        "depth_tor": 2,
        "depth_hyp": 7,
        "width": 15,
    },
]

CELLS = dict(
    mag="B24",
    dist_rup="B27",
    dist_jb="B30",
    dist_x="B33",
    dist_y0="B36",
    v_s30="B39",
    flag_u="B42",
    flag_rv="B45",
    flag_nm="B48",
    flag_hw="B51",
    dip="B54",
    depth_tor="B57",
    depth_hyp="B60",
    depth_1_0="B63",
    depth_2_5="B66",
    width="B69",
    flag_meas="B72",
    flag_as="B75",
    region="B78",
    dpp_centered="B83",
    depth_bor="B89",
)

RESULT_COLUMNS = dict(
    ASK14="DEF",
    BSSA14="DST",
    CB14="DYZ",
    CY14="DUV",
    I14="DLM",
)

abbreviations = ["ASK14", "BSSA14", "CB14", "CY14", "I14"]

# Join the parameters of the site and fault
parameters = dict(params_site)
for pf in params_fault:
    for k, v in pf.items():
        if k not in parameters:
            parameters[k] = list()
        parameters[k].append(v)

tests = []
for p in iter_parameters(parameters):
    wb = xw.Workbook(wb_fname)
    load_params(wb, **p)

    results = {abb: get_results(wb, abb) for abb in abbreviations}

    if p["region"] == "taiwan":
        # Load ASK14 specific values.
        load_params(wb, region="TWA")
        results["ASK14"] = get_results(wb, "ASK14")

    tests.append({"params": p, "results": results})

fname = "../tests/data/ngaw2_tests.json.gz"
with gzip.open(fname, "wt") as fp:
    json.dump(tests, fp, indent=4)



#!/usr/bin/env python
"""Create test cases for the DBC13 model."""

import gzip
import json
import os

import xlwings as xw

params = {
    "dist": [5, 40, 200],
    "mag": [4, 6, 7],
    "mechanism": ["NS", "RS", "SS"],
    "v_s30": [200, 400, 750, 1200],
}


def iter_parameters(parameters):
    max_len = max(len(p) for p in parameters.values())

    for i in range(max_len):
        yield {k: v[i % len(v)] for k, v in parameters.items()}


def reformat_params(params):
    """Translate to spreadsheet values."""
    params["mechanism"] = ["NS", "RS", "SS"].index(params["mechanism"])
    return params


def load_params(wb, **params):
    cell_map = {
        "mag": "B1",
        "v_s30": "B6",
        "mechanism": "B5",
    }
    params = reformat_params(dict(params))
    for k, v in params.items():
        if k == "dist":
            # Update all distances
            for c in ["B2", "B3", "B4"]:
                xw.Range("Input-Output", c, wkb=wb).value = v
        else:
            xw.Range("Input-Output", cell_map[k], wkb=wb).value = v

    # Force a re-calculation. This is the same as Ctrl + Alt + F9
    wb.xl_app.CalculateFull()


def get_results(wb):
    results = []
    for col, dist in zip("GHI", ["jb", "epi", "hyp"]):
        d = {}
        # Collect PGA and PGV
        for key, row in zip(["pga", "pgv"], [3, 4]):
            d[key] = xw.Range("Input-Output", "%s%d" % (col, row), wkb=wb).value

        # Collect response spectrum
        for key, c in zip(["periods", "spec_accels"], ["F", col]):
            d[key] = xw.Range("Input-Output", f"{c}5:{c}66", wkb=wb).value

        results.append(("dist_" + dist, d))

    return results


wb_fname = os.path.abspath("ASB14.xlsx")
tests = []
for p in iter_parameters(params):
    wb = xw.Workbook(wb_fname)
    load_params(wb, **p)
    tests.append({"params": p, "results": get_results(wb)})

fname = "../tests/data/asb14_tests.json.gz"
with gzip.open(fname, "wt") as fp:
    json.dump(tests, fp, indent=4)



"""Abrahamson, Gregor, and Addo (2016) ground motion moodel."""

import numpy as np

from . import model

__author__ = "Albert Kottke"


class AbrahamsonGregorAddo2016(model.GroundMotionModel):
    """Abrahamson, Gregor, and Addo (2016) ground motion moodel.

    This model was developed for subduction regions and is commonly referred to
    as the BCHydro model.

    Parameters
    ----------
    scenario : :class:`pygmm.model.Scenario`
        earthquake scenario

    """

    NAME = "Abrahamson, Gregor, & Addo (2016)"
    ABBREV = "AGA16"

    # Reference shear-wave velocity in m/sec
    V_REF = 1000.0

    # Load the coefficients for the model
    COEFF = model.load_data_file("abrahamson_gregor_addo_2016.csv", 1)
    PERIODS = COEFF["period"]

    INDEX_PGA = 0
    INDICES_PSA = np.arange(1, 23)

    # FIXME
    LIMITS = dict(
        mag=(3.0, 8.5),
        dist_jb=(0.0, 300.0),
        v_s30=(150.0, 1500.0),
    )

    PARAMS = [
        model.NumericParameter("mag", True, 3, 8.5),
        model.NumericParameter("dist_rup", False),
        model.NumericParameter("dist_hyp", False),
        model.NumericParameter("depth_hyp", False),
        model.NumericParameter("v_s30", True, 150.0, 1500.0),
        model.CategoricalParameter("event_type", True, ["interface", "intraslab"]),
        model.CategoricalParameter(
            "tectonic_region", False, ["forearc", "backarc", "unknown"], "unknown"
        ),
    ]

    def __init__(self, scenario, adjust_c1=None, adjust_c4=0, scale_atten=1.0):
        """Initialize the model.

        Args:
            scenario (:class:`pygmm.model.Scenario`): earthquake scenario.
        """
        super().__init__(scenario)

        n = len(self.COEFF)
        if adjust_c1 is None:
            # Default adjustments
            if self.scenario.event_type == "intraslab":
                self._adjust_c1 = -0.3 * np.ones(n)
            else:
                self._adjust_c1 = np.interp(
                    np.log(np.maximum(0.01, self.COEFF.period)),
                    np.log([0.3, 0.5, 1, 2, 3]),
                    [0.2, 0.1, 0, -0.1, -0.2],
                    left=0.2,
                    right=-0.2,
                )
        else:
            if isinstance(adjust_c1, float):
                self._adjust_c1 = adjust_c1 * np.ones(n)
            else:
                self._adjust_c1 = np.asarray(adjust_c1)

        self._adjust_c4 = adjust_c4
        self._scale_atten = scale_atten

        pga_ref = np.exp(self._calc_ln_resp(np.nan)[self.INDEX_PGA])
        self._ln_resp = self._calc_ln_resp(pga_ref)
        self._ln_std = self._calc_ln_std()

    @property
    def adjust_c1(self):
        return self._adjust_c1

    @property
    def adjust_c4(self):
        return self._adjust_c4

    @property
    def scale_atten(self):
        return self._scale_atten

    def _calc_ln_resp(self, pga_ref):
        """Calculate the natural logarithm of the response.

        Parameters
        ----------
        pga_ref : float
            peak ground acceleration (g) at the reference
            condition. If :class:`np.nan`, then no site term is applied.

        Returns
        -------
        ln_resp : class:`np.array`:
            natural log of the response

        """
        s = self._scenario
        c = self.COEFF

        f_event = 1 if s.event_type == "intraslab" else 0
        dist = s.dist_hyp if s.event_type == "intraslab" else s.dist_rup

        path_atten = (c.t_2 + c.t_14 * f_event + c.t_3 * (s.mag - 7.8)) * np.log(
            dist + c.c_4 * np.exp(c.t_9 * (s.mag - 6))
        )

        ln_resp = (
            c.t_1
            + c.t_4 * self.adjust_c1
            + path_atten
            + self._scale_atten * c.t_6 * dist
            + c.t_10 * f_event
            + self._calc_f_mag(s.mag)
            + self._calc_f_site(pga_ref)
        )

        if s.tectonic_region == "backarc":
            ln_resp += self._calc_f_faba(dist)

        if s.event_type == "intraslab":
            ln_resp += self._calc_f_depth(s.depth_hyp)

        return ln_resp

    def _calc_f_mag(self, mag):
        """Calculate the magnitude scaling. Equation 2.

        Parameters
        ----------
        mag : float
            moment magnitude

        Returns
        -------
        f_mag : class:`np.array`:
            magnitude scaling

        """
        c = self.COEFF
        c_1 = 7.8
        coeff = np.select([mag <= (c_1 + self.adjust_c1), True], [c.t_4, c.t_5])
        f_mag = coeff * (mag - (c_1 + self.adjust_c1)) + c.t_13 * (10 - mag) ** 2
        return f_mag

    def _calc_f_depth(self, depth_hyp):
        """Calculate the depth scaling. Equation 3.

        Parameters
        ----------
        depth_hyp : float
            depth to hypocenter [km]

        Returns
        -------
        f_depth : class:`np.array`:
            depth scaling

        """
        c = self.COEFF
        f_depth = c.t_11 * (np.minimum(depth_hyp, 120) - 60)
        return f_depth

    def _calc_f_faba(self, dist):
        """Calculate the forearc/backarc scaling. Equation 4.

        Parameters
        ----------
        dist : float
            depth to hypocenter [km]

        Returns
        -------
        f_faba : class:`np.array`:
            forearc/backarc scaling

        """
        c = self.COEFF
        if self.scenario.event_type == "intraslab":
            f_faba = c.t_7 + c.t_8 * np.log(np.maximum(dist, 85) / 40)
        elif self.scenario.event_type == "interface":
            f_faba = c.t_15 + c.t_16 * np.log(np.maximum(dist, 100) / 40)
        else:
            raise NotImplementedError

        return f_faba

    def _calc_f_site(self, pga_ref):
        """Calculate the site amplification. Equation 5.

        Parameters
        ----------
        pga_ref : float, default: np.nan
            reference PGA (V_s30 = 1000 m/sec). If 'np.nan', then the ground
            motion at the reference condition is computed.

        Returns
        -------
        f_site : class:`np.array`:
            site scaling

        """
        c = self.COEFF
        v_s30 = 1000.0 if np.isnan(pga_ref) else self.scenario.v_s30
        vs_ratio = np.minimum(v_s30, 1000) / c.v_lin

        f_site = np.select(
            [v_s30 < c.v_lin, True],
            [
                c.t_12 * np.log(vs_ratio)
                - c.b * np.log(pga_ref + c.c)
                + c.b * np.log(pga_ref + c.c * vs_ratio**c.n),
                (c.t_12 + c.b * c.n) * np.log(vs_ratio),
            ],
        )
        return f_site

    def _calc_ln_std(self):
        """Calculate the logarithmic standard deviation.

        Returns
        -------
        ln_std : class:`np.array`:
            natural log standard deviation

        """
        c = self.COEFF

        ln_std = np.sqrt(c.phi**2 + c.tau**2)
        return ln_std



"""Abrahamson and Bhasin (2020, :cite:`abrahamson20`) model."""

from typing import Optional

import numpy as np

from . import model

__author__ = "Albert Kottke"


class AbrahamsonBhasin2020(model.Model):
    """Abrahamson (2020, :cite:`abrahamson20`) model.

    Conditional PGV model

    Parameters
    ----------
    scenario : :class:`pygmm.model.Scenario`
        earthquake scenario

    """

    NAME = "Abrahamson and Bhasin (2020)"
    ABBREV = "AB20"

    # Reference velocity (m/s)
    V_REF = 425.0

    # Load the coefficients for the model
    COEFFS = {
        # From Table 3.2
        "psa(T=Tpgv)": model.Coefficients(
            a_1=5.39,
            a_2=0.799,
            a_3=0.654,
            a_4=0.479,
            a_5=-0.062,
            a_6=-0.359,
            a_7=-0.134,
            a_8=0.023,
            phi=0.29,
            tau=0.16,
        ),
        "pga": model.Coefficients(
            a_1=4.77,
            a_2=0.738,
            a_3=0.484,
            a_4=0.275,
            a_5=-0.036,
            a_6=-0.332,
            a_7=-0.44,
            a_8=0,
            phi_1=0.32,
            phi_2=0.42,
            tau_1=0.12,
            tau_2=0.26,
            M_1=5.0,
            M_2=7.0,
        ),
        "psa(T=1s)": model.Coefficients(
            a_1=4.80,
            a_2=0.82,
            a_3=0.55,
            a_4=0.27,
            a_5=0.054,
            a_6=-0.382,
            a_7=-0.21,
            a_8=0.0,
            phi_1=0.28,
            phi_2=0.38,
            tau_1=0.12,
            tau_2=0.17,
            M_1=5.0,
            M_2=7.0,
        ),
    }

    PARAMS = [
        model.NumericParameter("dist_rup", True, 0, 300),
        model.NumericParameter("mag", True, 3.0, 8.0),
        model.NumericParameter("v_s30", True, 180, 1500),
    ]

    def __init__(
        self,
        scenario: model.Scenario,
        psa=Optional[float],
        pga=Optional[float],
        psa_1s=Optional[float],
    ):
        super().__init__(scenario)

        if psa is not None:
            method = "psa(T=Tpgv)"
            # FIXME: How to interpolate?
            raise NotImplementedError
        elif pga is not None:
            method = "pga"
            ln_gm = np.log(pga)
        else:
            method = "psa(T=1s)"
            ln_gm = np.log(psa_1s)

        C = self.COEFFS[method]
        s = scenario

        # Mean
        if s.mag < 5:
            f_1 = C.a_2
        elif s.mag > 7.5:
            f_1 = C.a_3
        else:
            f_1 = C.a_2 + (C.a_3 - C.a_2) * (s.mag - 5) / 2.5

        ln_mean = (
            C.a_1
            + f_1 * ln_gm
            + C.a_4 * (s.mag - 6)
            + C.a_5 * (8.5 - s.mag) ** 2
            + C.a_6 * np.log(s.dist_rup + 5 * np.exp(0.4 * (s.mag - 6)))
            + (C.a_7 + C.a_8 * (s.mag - 5)) * np.log(C.v_s30 / self.V_REF)
        )

        if method == "psa(T=Tpgv)":
            phi, tau = C.phi, C.tau
        else:

            def interp(var_1, var_2):
                if s.mag < C.mag_1:
                    val = var_1
                elif s.mag > C.mag_2:
                    val = var_2
                else:
                    val = var_1 + (var_2 - var_1) * (s.mag - C.mag_1) / (C.mag_2 - C.mag_1)
                return val

            phi = interp(C.phi_1, C.phi_2)
            tau = interp(C.tau_1, C.tau_2)

        ln_std = np.sqrt(phi**2 + tau**2)
        return ln_mean, ln_std

    @classmethod
    def ln_period_pgv(cls, mag: float) -> float:
        """Natural logarithm of the period for calculating PGV. Equation 3.6.

        Parameters
        ----------
        mag : float
            Magnitude

        Returns
        -------
        ln_period_pgv : float
            Best period for computing PGV.
        """
        return -4.09 + 0.66 * mag



#!/usr/bin/env python
"""Test calculation of ASK14 static methods."""

from numpy.testing import assert_allclose

from pygmm import AbrahamsonSilvaKamai2014 as ASK14


def test_depth_1_0():
    # Value calculated from NGAW2 spreadsheet
    assert_allclose(ASK14.calc_depth_1_0(600), 0.1424470, rtol=1e-5)


def test_depth_tor():
    # Value calculated from NGAW2 spreadsheet
    assert_allclose(ASK14.calc_depth_tor(6), 4.2545455)


def test_width():
    # Value calculated from NGAW2 spreadsheet
    assert_allclose(ASK14.calc_width(6, 50), 8.9125094)



"""Model for the Campbell and Bozorgnia (2014) ground motion model."""

import logging
from typing import Optional

import numpy as np

from . import model
from .chiou_youngs_2014 import ChiouYoungs2014 as CY14
from .types import ArrayLike

__author__ = "Albert Kottke"


class CampbellBozorgnia2014(model.GroundMotionModel):
    """Campbell and Bozorgnia (2014, :cite:`campbell14`) model.

    This model was developed for active tectonic regions as part of the
    NGA-West2 effort.
    """

    NAME = "Campbell & Bozorgnia (2014)"
    ABBREV = "CB14"

    # Reference velocity (m/sec)
    V_REF = 1100.0

    # Load the coefficients for the model
    COEFF = model.load_data_file("campbell_bozorgnia_2014.csv", 2)

    PERIODS = COEFF["period"]

    # Period independent model coefficients
    COEFF_C = 1.88
    COEFF_N = 1.18
    COEEF_H_4 = 1

    INDICES_PSA = np.arange(21)
    INDEX_PGA = -2
    INDEX_PGV = -1

    PARAMS = [
        model.NumericParameter("depth_1_0", False),
        model.NumericParameter("depth_2_5", False, 0, 10),
        model.NumericParameter("depth_bor", False),
        model.NumericParameter("depth_bot", False, default=15.0),
        model.NumericParameter("depth_hyp", False, 0, 20),
        model.NumericParameter("depth_tor", False, 0, 20),
        model.NumericParameter("dip", True, 15, 90),
        model.NumericParameter("dist_jb", True),
        model.NumericParameter("dist_rup", True, None, 300),
        model.NumericParameter("dist_x", True),
        model.NumericParameter("mag", True, 3.3, 8.5),
        model.NumericParameter("v_s30", True, 150, 1500),
        model.NumericParameter("width", False),
        model.CategoricalParameter(
            "region",
            False,
            ["global", "california", "japan", "italy", "china"],
            "global",
        ),
        model.CategoricalParameter("mechanism", True, ["SS", "NS", "RS"]),
    ]

    def _check_inputs(self) -> None:
        """Check the inputs."""
        super()._check_inputs()
        s = self._scenario

        for mech, limit in [("SS", 8.5), ("RS", 8.0), ("NS", 7.5)]:
            if mech == s.mechanism and s.mag > limit:
                logging.warning(
                    "Magnitude of %g is greater than the recommended limit of"
                    "%g for %s style faults",
                    s.mag,
                    limit,
                    mech,
                )
        if s.depth_2_5 is None:
            s.depth_2_5 = self.calc_depth_2_5(s.v_s30, s.region, s.depth_1_0)

        if s.depth_tor is None:
            s.depth_tor = CY14.calc_depth_tor(s.mag, s.mechanism)

        if s.width is None:
            s.width = CampbellBozorgnia2014.calc_width(s.mag, s.dip, s.depth_tor, s.depth_bot)

        if s.depth_bor is None:
            s.depth_bor = self.calc_depth_bor(s.depth_tor, s.dip, s.width)

        if s.depth_hyp is None:
            s.depth_hyp = CampbellBozorgnia2014.calc_depth_hyp(
                s.mag, s.dip, s.depth_tor, s.depth_bor
            )

    def __init__(self, scenario: model.Scenario):
        """Initialize the model.

        Args:
            scenario (:class:`pygmm.model.Scenario`): earthquake scenario.
        """
        super().__init__(scenario)

        pga_ref = np.exp(self._calc_ln_resp(np.nan, self.V_REF)[self.INDEX_PGA])
        self._ln_resp = self._calc_ln_resp(pga_ref, self._scenario.v_s30)
        self._ln_std, self._tau, self._phi = self._calc_ln_std(pga_ref)

    def _calc_ln_resp(self, pga_ref: float, v_s30: float) -> np.ndarray:
        """Calculate the natural logarithm of the response.

        Parameters
        ----------
        pga_ref : float
            peak ground acceleration (g) at the reference
            condition. If :class:`np.nan`, then no site term is applied.
        v_s30 : float
            time-averaged shear-wave velocity over the top 30 m
            of the site (:math:`V_{s30}`, m/s).

        Returns
        -------

            class:`np.array`: Natural log of the response.

        """
        c = self.COEFF
        s = self._scenario

        # Magnitude term
        f_mag = c.c_0 + c.c_1 * s.mag
        for min_mag, slope in ([4.5, c.c_2], [5.5, c.c_3], [6.5, c.c_4]):
            if min_mag < s.mag:
                f_mag += slope * (s.mag - min_mag)
            else:
                break

        # Geometric attenuation term
        f_dis = (c.c_5 + c.c_6 * s.mag) * np.log(np.sqrt(s.dist_rup**2 + c.c_7**2))

        # Style of faulting term
        taper = np.clip(s.mag - 4.5, 0, 1)
        if s.mechanism == "RS":
            f_flt = c.c_8 * taper
        elif s.mechanism == "NS":
            f_flt = c.c_9 * taper
        else:
            f_flt = 0

        # Hanging-wall term
        R_1 = s.width * np.cos(np.radians(s.dip))
        R_2 = 62 * s.mag - 350
        if s.dist_x < 0:
            f_hngRx = 0
        elif s.dist_x <= R_1:
            ratio = s.dist_x / R_1
            f_hngRx = c.h_1 + c.h_2 * ratio + c.h_3 * ratio**2
        else:
            ratio = (s.dist_x - R_1) / (R_2 - R_1)
            f_hngRx = np.maximum(0, c.h_4 + c.h_5 * ratio + c.h_6 * ratio**2)

        if s.dist_rup == 0:
            f_hngRrup = 1
        else:
            f_hngRrup = (s.dist_rup - s.dist_jb) / s.dist_rup

        if s.mag <= 5.5:
            f_hngM = 0
        else:
            f_hngM = np.minimum(s.mag - 5.5, 1) * (1 + c.a_2 * (s.mag - 6.5))

        f_hngZ = 0 if s.depth_tor > 16.66 else 1 - 0.06 * s.depth_tor
        f_hngDip = (90 - s.dip) / 45

        f_hng = c.c_10 * f_hngRx * f_hngRrup * f_hngM * f_hngZ * f_hngDip

        site_term = self.calc_site_term(pga_ref, v_s30, s.depth_2_5, s.region)

        # Hypocentral depth term
        f_hypH = np.clip(s.depth_hyp - 7, 0, 13)
        f_hypM = c.c_17 + (c.c_18 - c.c_17) * np.clip(s.mag - 5.5, 0, 1)
        f_hyp = f_hypH * f_hypM

        # Fault dip term
        f_dip = c.c_19 * s.dip * np.clip(5.5 - s.mag, 0, 1)

        # Anaelastic attenuation term
        if s.region in ["japan", "italy"]:
            dc_20 = c.dc_20jp
        elif s.region == ["china"]:
            dc_20 = c.dc_20ch
        else:
            dc_20 = c.dc_20ca

        f_atn = (c.c_20 + dc_20) * max(s.dist_rup - 80, 0)

        ln_resp = f_mag + f_dis + f_flt + f_hng + site_term + f_hyp + f_dip + f_atn
        return ln_resp

    @classmethod
    def calc_site_term(
        cls,
        pga_ref: float,
        v_s30: float,
        depth_2_5: float,
        region: str = "california",
    ) -> ArrayLike:
        """Calculate the site term, which includes site and basin effects.

        Parameters
        ----------
        resp_ref :  array_like, optional
            response at the reference condition
        v_s30 : float
            site condition. Set `v_s30` to the reference
            velocity (e.g., 1180 m/s) for the reference response.
        depth_2_5 : float
            depth to the 2.5 km∕s shear-wave velocity horizon beneath the site,
            :math:`Z_{2.5}` in (km).
        region : str, optional
            region of basin model. Valid options:
            "global", "california", "japan", "italy", "china". If
            *None*, then "global" is used as the default value.

        Returns
        -------
        site_term: :class:`np.ndarray`
            site term that is applied to the natural log response.
        """

        c = cls.COEFF

        # Site term
        f_site = np.zeros_like(c.period)
        vs_ratio = v_s30 / c.k_1
        mask = v_s30 <= c.k_1
        f_site[mask] = (
            c.c_11 * np.log(vs_ratio)
            + c.k_2
            * (
                np.log(pga_ref + cls.COEFF_C * vs_ratio**cls.COEFF_N)
                - np.log(pga_ref + cls.COEFF_C)
            )
        )[mask]
        f_site[~mask] = ((c.c_11 + c.k_2 * cls.COEFF_N) * np.log(vs_ratio))[~mask]

        if region == "japan":
            # Apply regional correction for Japan
            if v_s30 <= 200:
                f_site += (c.c_12 + c.k_2 * cls.COEFF_N) * (np.log(vs_ratio) - np.log(200 / c.k_1))
            else:
                f_site += (c.c_13 + c.k_2 * cls.COEFF_N) * np.log(vs_ratio)

        # Basin response term
        if np.isnan(pga_ref):
            # Use model to compute depth_2_5 for the reference velocity case
            depth_2_5 = cls.calc_depth_2_5(v_s30, region)
        else:
            depth_2_5 = depth_2_5

        if depth_2_5 <= 1:
            f_sed = c.c_14 * (depth_2_5 - 1)
            if region == "japan":
                f_sed += c.c_15 * (depth_2_5 - 1)
        elif depth_2_5 <= 3:
            f_sed = 0
        else:
            f_sed = c.c_16 * c.k_3 * np.exp(-0.75) * (1 - np.exp(-0.25 * (depth_2_5 - 3)))

        return f_site + f_sed

    def _calc_ln_std(self, pga_ref: ArrayLike) -> (np.ndarray, np.ndarray, np.ndarray):
        """Calculate the logarithmic standard deviation.

        Parameters
        ----------
        pga_ref : float
            peak ground acceleration (g) at the reference
            condition.

        Returns
        -------

            class:`np.array`: Logarithmic standard deviation.

        """
        c = self.COEFF
        s = self._scenario

        tau_lnY = c.tau_2 + (c.tau_1 - c.tau_2) * np.clip(5.5 - s.mag, 0, 1)
        phi_lnY = c.phi_2 + (c.phi_1 - c.phi_2) * np.clip(5.5 - s.mag, 0, 1)

        vs_ratio = s.v_s30 / c.k_1
        alpha = np.zeros_like(c.period)
        mask = s.v_s30 < c.k_1
        alpha[mask] = (
            c.k_2
            * pga_ref
            * (
                (pga_ref + self.COEFF_C * vs_ratio**self.COEFF_N) ** (-1)
                - (pga_ref + self.COEFF_C) ** -1
            )
        )[mask]

        tau_lnPGA = tau_lnY[self.INDEX_PGA]
        tau = np.sqrt(
            tau_lnY**2 + alpha**2 * tau_lnPGA**2 + 2 * alpha * c.rho_lnPGAlnY * tau_lnY * tau_lnPGA
        )

        phi_lnPGA = phi_lnY[self.INDEX_PGA]
        phi_lnAF_PGA = self.COEFF["phi_lnAF"][self.INDEX_PGA]
        phi_lnPGA_B = np.sqrt(phi_lnPGA**2 - phi_lnAF_PGA**2)
        phi_lnY_B = np.sqrt(phi_lnY**2 - c.phi_lnAF**2)

        phi = np.sqrt(
            phi_lnY_B**2
            + c.phi_lnAF**2
            + alpha**2 * (phi_lnPGA**2 - phi_lnAF_PGA**2)
            + 2 * alpha * c.rho_lnPGAlnY * phi_lnY_B * phi_lnPGA_B
        )

        ln_std = np.sqrt(phi**2 + tau**2)

        return ln_std, tau, phi

    @staticmethod
    def calc_depth_2_5(
        v_s30: float, region: str = "global", depth_1_0: Optional[float] = None
    ) -> float:
        """Calculate the depth to a shear-wave velocity of 2.5 km/sec
        (:math:`Z_{2.5}`).

        Provide either `v_s30` or `depth_1_0`.

        Parameters
        ----------
        v_s30 : Optional[float]
            time-averaged shear-wave velocity over
            the top 30 m of the site (:math:`V_{s30}`, m/s).
            Keyword Args:
        region : Optional[str]
            region of the basin model. Valid values:
            "california", "japan". (Default value = 'global')
        depth_1_0 : Optional[float]
            depth to the 1.0 km∕s shear-wave
            velocity horizon beneath the site, :math:`Z_{1.0}` in (km).
            (Default value = None)

        Returns
        -------
        float
            estimated depth to a shear-wave velocity of 2.5 km/sec
        float
            estimated depth to a shear-wave velocity of 2.5 km/sec
            (km).

        """
        if v_s30:
            param = v_s30
            if region == "japan":
                # From equation 6.10 on page 63
                intercept = 5.359
                slope = 1.102
            else:
                # From equation 6.9 on page 63
                intercept = 7.089
                slope = 1.144

            # Global model
            # Not supported by NGA-West2 spreadsheet, and therefore removed.
            # foo = 6.510
            # bar = 1.181
        elif depth_1_0:
            param = depth_1_0
            if region == "japan":
                # From equation 6.13 on page 64
                intercept = 0.408
                slope = 1.745
            else:
                # From equation 6.12 on page 64
                intercept = 1.392
                slope = 1.798

            # Global model
            # Not supported by NGA-West2 spreadsheet, and therefore removed.
            # foo = 0.748
            # bar = 2.128
        else:
            raise NotImplementedError

        return np.exp(intercept - slope * np.log(param))

    @staticmethod
    def calc_depth_hyp(mag: float, dip: float, depth_tor: float, depth_bor: float) -> float:
        """Estimate the depth to hypocenter.

        Parameters
        ----------
        mag : float
            moment magnitude of the event (:math:`M_w`)
        dip : float
            fault dip angle (:math:`\\phi`, deg).
        depth_tor : float
            depth to the top of the rupture
            plane (:math:`Z_{tor}`, km).
        depth_bor : float
            depth to the bottom of the rupture
            plane (:math:`Z_{bor}`, km).

        Returns
        -------
        float
            estimated hypocenter depth (km)

        """
        # Equations 35, 36, and 37 of journal article
        ln_dZ = min(
            min(-4.317 + 0.984 * mag, 2.325) + min(0.0445 * (dip - 40), 0),
            np.log(0.9 * (depth_bor - depth_tor)),
        )

        depth_hyp = depth_tor + np.exp(ln_dZ)

        return depth_hyp

    @staticmethod
    def calc_width(mag: float, dip: float, depth_tor: float, depth_bot: float = 15.0) -> float:
        """Estimate the fault width using Equation (39) of CB14.

        Parameters
        ----------
        mag : float
            moment magnitude of the event (:math:`M_w`)
        dip : float
            fault dip angle (:math:`\\phi`, deg).
        depth_tor : float
            depth to the top of the rupture
            plane (:math:`Z_{tor}`, km).
            Keyword Args:
        depth_bot : Optional[float]
            depth to bottom of seismogenic crust
            (km). Used to calculate fault width if none is specified. If
            *None*, then a value of 15 km is used. (Default value = 15.0)

        Returns
        -------
        float
            estimated fault width (km)
        """
        return min(
            np.sqrt(10 ** ((mag - 4.07) / 0.98)),
            (depth_bot - depth_tor) / np.sin(np.radians(dip)),
        )

    @staticmethod
    def calc_depth_bor(depth_tor: float, dip: float, width: float) -> float:
        """Compute the depth to bottom of the rupture (km).

        Parameters
        ----------
        dip : float
            fault dip angle (:math:`\\phi`, deg).
        depth_tor : float
            depth to the top of the rupture
            plane (:math:`Z_{tor}`, km).
        width : float
            Down-dip width of the fault.

        Returns
        -------
        float
            depth to bottom of the fault rupture (km)
        """
        return depth_tor + width * np.sin(np.radians(dip))



"""Reference case from AbrahamsonSilva1996 based on spreadsheet by N.
Gregor."""

import json
import os

import pytest
from numpy.testing import assert_allclose

import pygmm

fpath = os.path.join(os.path.dirname(__file__), "data", "as96.json")
with open(fpath) as fp:
    data = json.load(fp)


@pytest.mark.parametrize("case", data)
def test_calc_duration(case):
    s = pygmm.Scenario(**case["scenario"])
    m = pygmm.AbrahamsonSilva1996(s)

    nias = case["nias"]
    stds = case["stds"]

    actual = m.interp(nias, stds)

    assert_allclose(actual, case["durs"], atol=0.0001, rtol=0.01)



#!/usr/bin/env python
"""Test model interface using C03 model."""

import pytest
from numpy.testing import assert_allclose, assert_array_equal

from pygmm import Campbell2003 as C03
from pygmm.model import Scenario


@pytest.fixture
def model():
    return C03(Scenario(mag=6.5, dist_rup=20))


def test_ln_std(model):
    assert_array_equal(model._ln_std, model.ln_stds)


def test_scenario():
    s = Scenario(mag=6, dist_jb=20)
    assert_allclose(s.mag, s["mag"])
    assert_allclose(s.dist_jb, s["dist_jb"])


@pytest.mark.parametrize("attr", ["pga", "ln_std_pga", "pgv", "ln_std_pgv", "pgd", "ln_std_pgd"])
def test_pga(model, attr):
    with pytest.raises(NotImplementedError):
        getattr(model, attr)



import json
import os

import numpy as np
import pytest
from numpy.testing import assert_allclose

from pygmm.kishida_2017 import calc_cond_mean_spectrum_vector

fpath = os.path.join(os.path.dirname(__file__), "data", "kishida_2017.json")
with open(fpath) as fp:
    data = json.load(fp)


def idfn(case):
    if isinstance(case, dict):
        return str(case["periods_cond"])


@pytest.mark.parametrize("case", data["cases"], ids=idfn)
@pytest.mark.parametrize("param", ["psas_cmsv", "ln_stds_cmsv"])
def test_calc_cond_mean_spectrum_vector(case, param):
    model = data["model"]
    ln_psas_cond = np.ma.array(
        np.log(model["psas_target"]),
        mask=(~np.in1d(model["periods"], case["periods_cond"])),
    )
    results = calc_cond_mean_spectrum_vector(
        model["periods"], np.log(model["psas"]), model["ln_stds"], ln_psas_cond
    )
    # Need to go from ln_psa to psa
    actual = np.exp(results[0]) if param == "psas_cmsv" else results[1]
    expected = case[param]
    assert_allclose(actual, expected, atol=0.0001, rtol=0.0005)



#!/usr/bin/python
import gzip
import json
import os

import numpy as np
import pytest

from pygmm import AkkarSandikkayaBommer2014 as ASB14
from pygmm.model import Scenario

# Relative tolerance for all tests
RTOL = 1e-2

# Load the tests
fname = os.path.join(os.path.dirname(__file__), "data", "asb14_tests.json.gz")
with gzip.open(fname, "rt") as fp:
    tests = json.load(fp)

testdata = [(dist, t["params"], results) for t in tests for dist, results in t["results"]]


def create_model(params, dist):
    params = dict(params)
    params[dist] = params.pop("dist")
    s = Scenario(**params)
    m = ASB14(s)
    return m


@pytest.mark.parametrize("dist,params,expected", testdata)
def test_spec_accels(dist, params, expected):
    m = create_model(params, dist)
    np.testing.assert_allclose(
        m.interp_spec_accels(expected["periods"]),
        expected["spec_accels"],
        rtol=RTOL,
        err_msg="Spectral accelerations",
    )


@pytest.mark.parametrize("dist,params,expected", testdata)
@pytest.mark.parametrize("key", ["pga", "pgv"])
def test_im_values(dist, params, expected, key):
    m = create_model(params, dist)
    np.testing.assert_allclose(
        getattr(m, key),
        expected[key],
        rtol=RTOL,
    )



"""Baker and Jayaram (2008, :cite:`baker08`) correlation model."""

import numpy as np

from .types import ArrayLike

__author__ = "Albert Kottke"


def calc_correls(periods: ArrayLike, period_cond: float) -> np.ndarray:
    """Baker and Jayaram (2008, :cite:`baker08`) correlation model.

    Parameters
    ----------
    periods : array_like
        Periods at which the correlation should be computed.
    period_cond : float
        Conditioning period

    Returns
    -------
    correls : :class:`np.ndarray`
        Correlation coefficients
    """
    period_min = np.minimum(periods, period_cond)
    period_max = np.maximum(periods, period_cond)

    c_1 = 1 - np.cos(np.pi / 2 - 0.366 * np.log(period_max / np.maximum(period_min, 0.109)))

    # The minimum() is added to prevent an overflow issue
    c_2 = np.select(
        [period_max < 0.2, True],
        [
            1
            - 0.105
            * (1 - 1 / (1 + np.exp(100 * np.minimum(period_max, 0.2) - 5)))
            * (period_max - period_min)
            / (period_max - 0.0099),
            0,
        ],
    )

    c_3 = np.select([period_max < 0.109, True], [c_2, c_1])

    c_4 = c_1 + 0.5 * (np.sqrt(c_3) - c_3) * (1 + np.cos(np.pi * period_min / 0.109))

    correls = np.select(
        [period_max < 0.109, period_min > 0.109, period_max < 0.200, True],
        [c_2, c_1, np.minimum(c_2, c_4), c_4],
    )

    return correls


def calc_cond_mean_spectrum(
    periods: ArrayLike,
    ln_psas: ArrayLike,
    ln_stds: ArrayLike,
    period_cond: float,
    ln_psa_cond: float,
) -> (np.ndarray, np.ndarray):
    """Conditional mean spectrum by Baker & Jayaram (2008, :cite:`baker08`).

    Parameters
    ----------
    periods : array_like
        Response spectral periods.
    ln_psas : array_like
        Natural logarithm of the 5%-damped spectral accelerations.
    ln_stds : array_like
        Logarithmic standard deviations.
    period_cond : float
        Conditioning period. This period does not need to be included in
        `periods`.
    ln_psa_cond : float
        Natural logarithm of the response at the conditioning period.

    Returns
    -------
    ln_psas_cms : :class:`np.ndarray`
        Natural logarithm of the conditional mean spectral accelerations.
    ln_stds_cms : :class:`np.ndarray`
        Logarithmic standard deviation of the conditional mean spectral
        acceleration.
    """
    periods = np.asarray(periods)
    ln_psas = np.asarray(ln_psas)
    ln_stds = np.asarray(ln_stds)

    correls = calc_correls(periods, period_cond)
    epsilon = (ln_psa_cond - np.interp(period_cond, periods, ln_psas)) / np.interp(
        period_cond, periods, ln_stds
    )

    ln_psas_cms = ln_psas + ln_stds * correls * epsilon
    ln_stds_cms = np.sqrt(ln_stds**2 * (1 - correls**2))
    return ln_psas_cms, ln_stds_cms



#!/usr/bin/env python
"""Create plots for visual comparison."""

import os

import matplotlib

matplotlib.use("agg")  # NOQA
import matplotlib.pyplot as plt
import pytest

import pygmm

DEFAULT_PROPS = dict(
    depth_2_5=5,
    depth_bor=15,
    depth_hyp=9,
    depth_tor=5,
    dip=90.0,
    dist=20,
    dist_jb=30.0,
    dist_rup=30.0,
    dist_x=30.0,
    dpp_centered=0,
    flag_hw=0,
    flag_meas=0,
    mag=6,
    mechanism="SS",
    region="california",
    v_s30=500.0,
    width=10,
)

# Make the figure directory if needed
if not os.path.exists("figures"):
    os.makedirs("figures")


@pytest.mark.parametrize("model", pygmm.models, ids=lambda m: m.ABBREV)
@pytest.mark.parametrize(
    "key,values,label",
    [
        ("mag", [5, 6, 7], "Magnitude"),
        (["dist", "dist_rup", "dist_jb", "dist_x"], [10, 50, 100], "Distance (km)"),
        ("v_s30", [300, 650, 1000], "$V_{s30}$ (m/s)"),
    ],
    ids=lambda a: a[0],
)
def plot_model_with_param(model, key, values, label):
    props = dict(DEFAULT_PROPS)

    fig, ax = plt.subplots()
    for v in values:
        if isinstance(key, str):
            props[key] = v
        else:
            for k in key:
                props[k] = v
        m = model(**props)
        ax.plot(m.periods, m.spec_accels, label="%g" % v)

    ax.set_xlabel("Period (s)")
    try:
        ax.set_xscale("log")
    except ValueError:
        pass

    ax.set_ylabel("5% Damped, Spectral. Accel. (g)")
    ax.set_yscale("log")
    ax.set_ylim(1e-4, 1e1)

    ax.legend(loc="upper right", title=label, fontsize="x-small")

    ax.grid()

    fig.tight_layout()

    if isinstance(key, str):
        prefix = key
    else:
        prefix = key[0]

    fig.savefig(os.path.join("figures", prefix + "-" + model.ABBREV))
    plt.close(fig)



import numpy as np
import pytest

from pygmm import GulerceAbrahamson2011 as GA11

from . import load_tests

TESTS = load_tests("gulerce_abrahamson_2011.json.gz")


def idfn(test):
    if isinstance(test, dict):
        p = test["params"]
        return f"M={p['mag']}, Rjb={['dist_rup']} km"
    else:
        return test


@pytest.mark.parametrize("test", TESTS, ids=idfn)
@pytest.mark.parametrize("key", ["ratio", "ln_std"])
def test_calc_cond_spectrum(test, key):
    m = GA11(test["params"])
    actual = getattr(m, key)
    expected = test["output"][key]
    np.testing.assert_allclose(actual, expected, atol=0.001, rtol=0.005)


@pytest.mark.parametrize("method", ["interp_ratio", "interp_ln_std"])
def test_interp(method):
    m = GA11(TESTS[0]["params"])
    getattr(m, method)(np.geomspace(0.01, 10))



"""Bayless and Abrahamson (2018, :cite:`bayless19`) correlation."""

from typing import Optional

import numpy as np
import numpy.typing as npt

from . import model

__author__ = "Albert Kottke"


class BaylessAbrahamson2018:
    """Bayless and Abrahamson (2018) model.

    This inter-frequency correlation model was developed for active tectonic regions.
    """

    #: Long name of the model
    NAME = "Bayless and Abrahamson (2018)"
    #: Short name of the model
    ABBREV = "BA18"

    # Load the coefficients for the model
    COEFF = model.load_data_file("bayless_abrahamson_2018.csv", 3)
    FREQS = COEFF["freq_hz"]

    @classmethod
    def corr(cls, freqs: npt.ArrayLike) -> np.ndarray:
        """
        Parameters
        ----------
        freq : array_like
            Frequencies

        Returns
        -------
        corr : np.ndarray
            Correlation coefficient matrix
        """

        # Create two matrices of frequencies
        n = freqs.shape[0]
        triu = np.triu(np.full((n, n), freqs))
        freqs_mat = triu + triu.T - np.diag(freqs)

        def do_interp(f_m, attr):
            fp = getattr(cls.COEFF, attr)
            return np.interp(f_m, cls.FREQS, fp, left=fp[0], right=fp[-1])

        # Compute the frequency parameters
        f_r = np.abs(np.log(freqs / freqs_mat))
        f_m = np.minimum(freqs, freqs_mat)
        a, b, c, d = (do_interp(f_m, param) for param in "ABCD")
        # Lower matrix
        lower = np.tanh(a * np.exp(b * f_r) + c * np.exp(d * f_r)).reshape(n, n)

        corr = lower + lower.T
        np.fill_diagonal(corr, 1)

        return corr

    @classmethod
    def cov(
        cls, freqs, *, std: Optional[npt.ArrayLike] = None, component: Optional[str] = None
    ) -> np.ndarray:
        """
        Parameters
        ----------
        freq : array_like
            Frequencies [Hz]
        std : array_like, optional
            Standard deviation at the frequencies
        component : str, optional
            Component of the published standard deviation. Possible options: tau, phi_s2s, phi_s, sigma
        Returns
        -------
        corr : np.ndarray
            Correlation coefficient matrix
        """

        corr = cls.corr(freqs)
        if std is None and component is not None:
            fp = getattr(cls.COEFF, component)
            std = np.interp(freqs, cls.FREQS, fp, left=fp[0], right=fp[-1])
        elif std is None and component is None:
            raise NotImplementedError

        n = len(freqs)
        std = np.full((n, n), std)
        return corr * std * std.T



#!/usr/bin/python
import gzip
import json
import os

import numpy as np
import pytest

from pygmm import DerrasBardCotton2014 as DBC14
from pygmm.model import Scenario

# Relative tolerance for all tests
RTOL = 1e-2

# Load the tests
fname = os.path.join(os.path.dirname(__file__), "data", "dbc14_tests.json.gz")
with gzip.open(fname, "rt") as fp:
    tests = json.load(fp)

testdata = [(t["params"], t["results"]) for t in tests]


@pytest.mark.parametrize("params,expected", testdata)
def test_spec_accels(params, expected):
    m = DBC14(Scenario(**params))
    np.testing.assert_allclose(
        m.interp_spec_accels(expected["periods"]),
        # Need to convert from m/sec to g
        np.array(expected["spec_accels"]) / m.GRAVITY,
        rtol=RTOL,
        err_msg="Spectral accelerations",
    )


@pytest.mark.parametrize("params,expected", testdata)
@pytest.mark.parametrize("key", ["pga", "pgv"])
def test_im_values(params, expected, key):
    m = DBC14(Scenario(**params))
    # PGA needs to be converted from m/sec² to g, and PGV need to be
    # converted from m/sec into cm/sec
    scale = m.GRAVITY if key == "pga" else 0.01

    np.testing.assert_allclose(
        getattr(m, key),
        expected[key] / scale,
        rtol=RTOL,
    )


@pytest.fixture
def model():
    """Instance of the DBC13 model."""
    return DBC14(Scenario(dist_jb=10, mag=6, v_s30=600, depth_hyp=10, mechanism="SS"))


def test_ln_std(model):
    # Log10 total standard deviations from the paper
    expected = [
        0.298,
        0.309,
        0.31,
        0.313,
        0.319,
        0.323,
        0.325,
        0.328,
        0.335,
        0.338,
        0.338,
        0.337,
        0.338,
        0.337,
        0.335,
        0.334,
        0.333,
        0.33,
        0.328,
        0.328,
        0.326,
        0.325,
        0.322,
        0.322,
        0.326,
        0.328,
        0.329,
        0.33,
        0.33,
        0.329,
        0.327,
        0.328,
        0.328,
        0.327,
        0.33,
        0.331,
        0.332,
        0.332,
        0.332,
        0.331,
        0.33,
        0.331,
        0.333,
        0.335,
        0.339,
        0.343,
        0.346,
        0.353,
        0.356,
        0.359,
        0.362,
        0.365,
        0.368,
        0.368,
        0.37,
        0.37,
        0.373,
        0.375,
        0.377,
        0.378,
        0.378,
        0.376,
        0.375,
        0.375,
    ]

    np.testing.assert_allclose(np.log10(np.exp(model._ln_std)), expected, rtol=RTOL)



"""Kempton and Stewart (2006, :cite:`kempton2006`) duration model."""

import numpy as np

from . import model

__author__ = ""


class KemptonStewart2006(model.Model):
    """Kempton and Stewart (2006, :cite:`kempton2006`) duration model.

    Parameters
    ----------
    scenario : :class:`pygmm.model.Scenario`
        earthquake scenario

    """

    NAME = "Kempton Stewart (2006)"
    ABBREV = "KS06"

    PARAMS = [
        model.NumericParameter("mag", True, 5, 7.6),
        model.NumericParameter("dist_rup", True, 0, 200),
        model.NumericParameter("v_s30", True, 200, 1000),
    ]

    def __init__(self, scenario):
        super().__init__(scenario)

        # scenario
        s = self._scenario

        # stress drop indices for duration measures
        stress_drop = np.exp(
            [
                6.02,  # D5-75a
                2.79 + 0.82 * (s.mag - 6),  # D5-95a
                5.46,  # D5-75v
                1.53 + 1.34 * (s.mag - 6),
            ]
        )  # D5-95v

        # source duration
        f_0 = self._source_dur(stress_drop)

        # path duration
        f_1 = np.array([0.07, 0.15, 0.10, 0.15]) * s.dist_rup

        # site duration
        f_2 = (
            np.array([0.82, 3.00, 1.40, 3.99])
            + np.array([-0.0013, -0.0041, -0.0022, -0.0062]) * s.v_s30
        )

        # total druation
        self._ln_dur = np.log(f_0 + f_1 + f_2)

        # aleatory standard deviation
        self._std_err = np.array([0.57, 0.47, 0.66, 0.52])

    @property
    def duration(self):
        return KemptonStewart2006._as_recarray(np.exp(self._ln_dur))

    @property
    def std_err(self):
        return KemptonStewart2006._as_recarray(self._std_err)

    @staticmethod
    def _as_recarray(values):
        return np.rec.fromarrays(
            values,
            names=[
                "D_5t75a",
                "D_5t95a",
                "D_5t75v",
                "D_5t95v",
            ],
        )

    def _source_dur(self, stress_drop):
        # seismic moment
        moment = 10 ** (1.5 * self.scenario.mag + 16.05)
        return (stress_drop / moment) ** (-1 / 3) / (4.9e6 * 3.2)



"""pyGMM: Ground motion models implemented in Python."""

import logging
from importlib.metadata import version

from .abrahamson_bhasin_2020 import AbrahamsonBhasin2020
from .abrahamson_gregor_addo_2016 import AbrahamsonGregorAddo2016
from .abrahamson_silva_1996 import AbrahamsonSilva1996
from .abrahamson_silva_kamai_2014 import AbrahamsonSilvaKamai2014
from .afshari_stewart_2016 import AfshariStewart2016
from .akkar_sandikkaya_bommer_2014 import AkkarSandikkayaBommer2014
from .atkinson_boore_2006 import AtkinsonBoore2006
from .bayless_abrahamson_2018 import BaylessAbrahamson2018
from .bayless_abrahamson_2019 import BaylessAbrahamson2019
from .boore_stewart_seyhan_atkinson_2014 import BooreStewartSeyhanAtkinson2014
from .campbell_2003 import Campbell2003
from .campbell_bozorgnia_2014 import CampbellBozorgnia2014
from .chiou_youngs_2014 import ChiouYoungs2014
from .coppersmith_bommer_2014 import CoppersmithBommer2014
from .derras_bard_cotton_2014 import DerrasBardCotton2014
from .gulerce_abrahamson_2011 import GulerceAbrahamson2011
from .idriss_2014 import Idriss2014
from .kempton_stewart_2006 import KemptonStewart2006
from .model import Scenario
from .pezeshk_zandieh_tavakoli_2011 import PezeshkZandiehTavakoli2011
from .tavakoli_pezeshk_2005 import TavakoliPezeshk05

__all__ = [
    "Scenario",
    "AbrahamsonBhasin2020",
    "AbrahamsonSilvaKamai2014",
    "AbrahamsonGregorAddo2016",
    "AfshariStewart2016",
    "AkkarSandikkayaBommer2014",
    "AtkinsonBoore2006",
    "BaylessAbrahamson2018",
    "BaylessAbrahamson2019",
    "BooreStewartSeyhanAtkinson2014",
    "Campbell2003",
    "CampbellBozorgnia2014",
    "ChiouYoungs2014",
    "CoppersmithBommer2014",
    "DerrasBardCotton2014",
    "GulerceAbrahamson2011",
    "KemptonStewart2006",
    "Idriss2014",
    "PezeshkZandiehTavakoli2011",
    "TavakoliPezeshk05",
]

__author__ = "Albert Kottke"
__copyright__ = "Copyright 2016 Albert Kottke"
__license__ = "MIT"
__title__ = "pyGMM"
__version__ = version("pyGMM")

# Set default logging handler to avoid "No handler found" warnings.
try:  # Python 2.7+
    from logging import NullHandler
except ImportError:

    class NullHandler(logging.Handler):
        def emit(self, record):
            pass


logging.getLogger(__name__).addHandler(NullHandler())

models = [
    AbrahamsonSilva1996,
    AbrahamsonSilvaKamai2014,
    AfshariStewart2016,
    AkkarSandikkayaBommer2014,
    AtkinsonBoore2006,
    BaylessAbrahamson2019,
    BooreStewartSeyhanAtkinson2014,
    Campbell2003,
    CampbellBozorgnia2014,
    ChiouYoungs2014,
    CoppersmithBommer2014,
    DerrasBardCotton2014,
    GulerceAbrahamson2011,
    KemptonStewart2006,
    Idriss2014,
    PezeshkZandiehTavakoli2011,
    TavakoliPezeshk05,
]



"""Chiou and Youngs (2014, :cite:`chiou14`) model."""

import logging
from typing import Optional

import numpy as np

from . import model
from .types import ArrayLike

__author__ = "Albert Kottke"


class ChiouYoungs2014(model.GroundMotionModel):
    """Chiou and Youngs (2014, :cite:`chiou14`) model.

    This model was developed for active tectonic regions as part of the
    NGA-West2 effort.

    Parameters
    ----------
    scenario : :class:`pygmm.model.Scenario`
        earthquake scenario

    """

    NAME = "Chiou and Youngs (2014)"
    ABBREV = "CY14"

    # Reference velocity (m/s)
    V_REF = 1130.0

    # Load the coefficients for the model
    COEFF = model.load_data_file("chiou_youngs_2014.csv", 2)

    PERIODS = COEFF["period"]

    INDICES_PSA = np.arange(24)
    INDEX_PGA = 24
    INDEX_PGV = 25

    PARAMS = [
        model.NumericParameter("dist_rup", True, 0, 300),
        model.NumericParameter("dist_x", True),
        model.NumericParameter("dist_jb", True),
        model.NumericParameter("mag", True, 3.5, 8.5),
        model.NumericParameter("v_s30", True, 180, 1500),
        model.NumericParameter("depth_tor", False, 0, 20),
        model.NumericParameter("depth_1_0", False),
        model.NumericParameter("dpp_centered", False, default=0),
        model.NumericParameter("dip", True),
        model.CategoricalParameter("mechanism", False, ["U", "SS", "NS", "RS"], "U"),
        model.CategoricalParameter("on_hanging_wall", False, [True, False], False),
        model.CategoricalParameter(
            "region", False, ["california", "china", "italy", "japan"], "california"
        ),
        model.CategoricalParameter("vs_source", False, ["measured", "inferred"], "measured"),
    ]

    def __init__(self, scenario: model.Scenario):
        """Initialize the model."""
        super().__init__(scenario)
        ln_resp_ref = self._calc_ln_resp_ref()
        self._ln_resp = self._calc_ln_resp_site(ln_resp_ref)
        self._ln_std = self._calc_ln_std(np.exp(ln_resp_ref))

    def _calc_ln_resp_ref(self) -> np.ndarray:
        """Calculate the natural logarithm of the reference response.

        Returns
        -------
        ln_resp_ref : class:`np.array`:
            natural log of the response

        """
        c = self.COEFF
        s = self._scenario

        cosh_mag = np.cosh(2 * max(s.mag - 4.5, 0))
        ln_resp = np.array(c.c_1)

        if s.mechanism == "RS":
            # Reverse fault term
            ln_resp += c.c_1a + c.c_1c / cosh_mag
        elif s.mechanism == "NS":
            # Normal fault term
            ln_resp += c.c_1b + c.c_1d / cosh_mag

        # Magnitude scaling
        ln_resp += c.c_2 * (s.mag - 6)
        ln_resp += (c.c_2 - c.c_3) / c.c_n * np.log(1 + np.exp(c.c_n * (c.c_m - s.mag)))

        # Top of rupture term relative to model average
        diff_depth_tor = s.depth_tor - self.calc_depth_tor(s.mag, s.mechanism)
        ln_resp += (c.c_7 + c.c_7b / cosh_mag) * diff_depth_tor

        # Dip angle term
        ln_resp += (c.c_11 + c.c_11b / cosh_mag) * np.cos(np.radians(s.dip)) ** 2

        # Distance terms
        ln_resp += c.c_4 * np.log(
            s.dist_rup + c.c_5 * np.cosh(c.c_6 * np.maximum(s.mag - c.c_hm, 0))
        )
        ln_resp += (c.c_4a - c.c_4) * np.log(np.sqrt(s.dist_rup**2 + c.c_rb**2))

        # Regional adjustment
        if s.region in ["japan", "italy"] and (6 < s.mag < 6.9):
            scale = c.gamma_ji
        elif s.region in ["china"]:
            scale = c.gamma_c
        else:
            scale = 1.0
        ln_resp += (
            scale
            * (c.c_gamma1 + c.c_gamma2 / np.cosh(np.maximum(s.mag - c.c_gamma3, 0)))
            * s.dist_rup
        )

        # Directivity term
        ln_resp += (
            c.c_8
            * max(1 - max(s.dist_rup - 40, 0) / 30, 0)
            * min(max(s.mag - 5.5, 0) / 0.8, 1)
            * np.exp(-c.c_8a * (s.mag - c.c_8b) ** 2)
            * s.dpp_centered
        )

        # Hanging wall term
        if s.on_hanging_wall:
            ln_resp += (
                c.c_9
                * np.cos(np.radians(s.dip))
                * (c.c_9a + (1 - c.c_9a) * np.tanh(s.dist_x / c.c_9b))
                * (1 - np.sqrt(s.dist_jb**2 + s.depth_tor**2) / (s.dist_rup + 1))
            )

        return ln_resp

    def _calc_ln_resp_site(self, ln_resp_ref: np.ndarray) -> np.ndarray:
        """Calculate the natural logarithm of the response.

        Parameters
        ----------
        ln_resp_ref : :class:`np.array`
            natural logarithm of the response at the reference site condition
            at each of the periods specified by the model coefficients.

        Returns
        -------
        ln_resp_rsite : class:`np.ndarray`:
            natural log of the response including the site effects

        """
        s = self._scenario

        site_term = self.calc_site_term(np.exp(ln_resp_ref), s.v_s30, s.depth_1_0, s.region)
        return ln_resp_ref + site_term

    @classmethod
    def calc_site_term(
        cls, resp_ref: ArrayLike, v_s30: float, depth_1_0: float, region: Optional[str]
    ) -> ArrayLike:
        """Calculate the site term, which includes site and basin effects.

        Parameters
        ----------
        resp_ref : :class:`np.array`
            natural logarithm of the response at the reference site condition
            at each of the periods specified by the model coefficients.
        v_s30 : float
            site condition. Set `v_s30` to the reference
            velocity (e.g., 1180 m/s) for the reference response.
        depth_1_0 : float
            depth to the 1.0 km∕s shear-wave velocity horizon beneath the site,
            :math:`Z_{1.0}` in (km).
        region : str, optional
            region of basin model. Valid options: 'california', 'japan'. If
            *None*, then 'california' is used as the default value.

        Returns
        -------
        site_term: :class:`np.ndarray`
            site term that is applied to the natural log response.
        """

        c = cls.COEFF

        if region in ["japan"]:
            phi_1 = c.phi_1jp
            phi_5 = c.phi_5jp
            phi_6 = c.phi_6jp
        else:
            phi_1 = c.phi_1
            phi_5 = c.phi_5
            phi_6 = c.phi_6

        ln_resp_ref = np.log(resp_ref)

        # Linear scaling
        site_term = phi_1 * min(np.log(v_s30 / 1130.0), 0)
        # Nonlinear scaling
        site_term += (
            c.phi_2
            * (np.exp(c.phi_3 * (min(v_s30, 1130.0) - 360.0)) - np.exp(c.phi_3 * (1130.0 - 360.0)))
            * np.log((np.exp(ln_resp_ref) + c.phi_4) / c.phi_4)
        )
        # Basin model
        diff_depth_1_0 = 1000 * (depth_1_0 - cls.calc_depth_1_0(v_s30, region))
        site_term += phi_5 * (1 - np.exp(-diff_depth_1_0 / phi_6))

        return site_term

    def _calc_ln_std(self, resp_ref: np.ndarray) -> np.ndarray:
        """Calculate the logarithmic standard deviation.

        Parameters
        ----------
        ln_resp_ref : :class:`np.array`
            natural logarithm of the response at the reference site condition
            at each of the periods specified by the model coefficients.

        Returns
        -------
        ln_std : class:`np.array`:
            natural log standard deviation

        """
        c = self.COEFF
        s = self._scenario

        if s.region in ["japan"]:
            sigma_2 = c.sigma_2jp
        else:
            sigma_2 = c.sigma_2

        clipped_mag = np.clip(s.mag, 5.0, 6.5) - 5.0
        tau = c.tau_1 + (c.tau_2 - c.tau_1) / 1.5 * clipped_mag

        nl_0 = (
            c.phi_2
            * (
                np.exp(c.phi_3 * (min(s.v_s30, 1130.0) - 360.0))
                - np.exp(c.phi_3 * (1130.0 - 360.0))
            )
            * (resp_ref / (resp_ref + c.phi_4))
        )

        flag_meas = 1 if s.vs_source == "measured" else 0
        phi_nl = (c.sigma_1 + (sigma_2 - c.sigma_1) / 1.5 * clipped_mag) * np.sqrt(
            c.sigma_3 * (1 - flag_meas) + 0.7 * flag_meas + (1 + nl_0) ** 2
        )

        ln_std = np.sqrt((1 + nl_0) ** 2 * tau**2 + phi_nl**2)
        return ln_std

    def _check_inputs(self) -> None:
        """Check the inputs."""
        super()._check_inputs()
        s = self._scenario

        if s.mechanism in ["RS", "NS"]:
            _min, _max = 3.5, 8.0
        else:
            _min, _max = 3.5, 8.5

        if not (_min <= s.mag <= _max):
            logging.warning(
                "Magnitude (%g) exceeds recommended bounds (%g to %g)" " for a %s earthquake!",
                s.mag,
                _min,
                _max,
                s.mechanism,
            )

        if s.get("depth_tor", None) is None:
            s["depth_tor"] = self.calc_depth_tor(s.mag, s.mechanism)

        if s.get("depth_1_0", None) is None:
            # Calculate depth (m) and convert to (km)
            s["depth_1_0"] = self.calc_depth_1_0(s.v_s30, s.region)

    @staticmethod
    def calc_depth_1_0(v_s30: float, region: str) -> float:
        """Calculate the depth to 1 km/sec (:math:`Z_{1.0}`).

        Parameters
        ----------
        v_s30 : float
            time-averaged shear-wave velocity over the top 30 m
            of the site (:math:`V_{s30}`, m/s).
        region : str
            basin region. Valid options: "california", "japan"

        Returns
        -------
        depth_1_0 : float
            estimated depth to a shear-wave velocity of 1 km/sec (km)

        """
        if region in ["japan"]:
            # Japan
            power = 2
            v_ref = 412.39
            slope = -5.23 / power
        else:
            # Global
            power = 4
            v_ref = 570.94
            slope = -7.15 / power

        return (
            np.exp(slope * np.log((v_s30**power + v_ref**power) / (1360.0**power + v_ref**power)))
            / 1000
        )

    @staticmethod
    def calc_depth_tor(mag: float, mechanism: str) -> float:
        """Calculate an estimate of the depth to top of rupture (km).

        Parameters
        ----------
        mag : float
            moment magnitude of the event (:math:`M_w`)
        mechanism : str
            fault mechanism. Valid options: "U", "SS", "NS",
            "RS".

        Returns
        -------
        depth_tor : float
            estimated depth to top of rupture (km)

        """
        if mechanism == "RS":
            # Reverse and reverse-oblique faulting
            depth_tor = 2.704 - 1.226 * max(mag - 5.849, 0)
        else:
            # Combined strike-slip and normal faulting
            depth_tor = 2.673 - 1.136 * max(mag - 4.970, 0)

        return max(depth_tor, 0) ** 2



#!/usr/bin/env python
"""Create test cases for the DBC14 model."""

import gzip
import json
import os

import xlwings as xw

params = {
    "depth_hyp": [10, 0, 25],
    "dist_jb": [5, 40, 200],
    "mag": [4, 6, 7],
    "mechanism": ["NS", "RS", "SS"],
    "v_s30": [200, 400, 800],
}


def iter_parameters(parameters):
    max_len = max(len(p) for p in parameters.values())

    for i in range(max_len):
        yield {k: v[i % len(v)] for k, v in parameters.items()}


def reformat_params(params):
    """Translate to spreadsheet values."""
    params["mechanism"] = 1 + ["NS", "RS", "SS"].index(params["mechanism"])
    return params


def load_params(wb, **params):
    cell_map = {
        "dist_jb": "B08",
        "mag": "B10",
        "v_s30": "B12",
        "depth_hyp": "B14",
        "mechanism": "B16",
    }
    params = reformat_params(dict(params))
    for k, v in params.items():
        print("Setting:", k, cell_map[k], v)
        xw.Range("ANN(PGA)", cell_map[k], wkb=wb).value = v

    # Force a re-calculation. This is the same as Ctrl + Alt + F9
    wb.xl_app.CalculateFull()


def get_results(wb):
    d = {}
    # Collect PGA and PGV
    for key, row in zip(["pga", "pgv"], [13, 10]):
        d[key] = xw.Range("ANN(PGA)", "O%d" % row, wkb=wb).value

    # Collect response spectrum
    for key, rc in zip(["periods", "spec_accels"], "NO"):
        d[key] = xw.Range("ANN(PGA)", f"{rc}17:{rc}78", wkb=wb).value

    return d


wb_fname = os.path.abspath("10518_2013_9481_MOESM1_ESM.xlsx")
tests = []
for p in iter_parameters(params):
    wb = xw.Workbook(wb_fname)
    load_params(wb, **p)
    tests.append({"params": p, "results": get_results(wb)})

fname = "../tests/data/dbc14_tests.json.gz"
with gzip.open(fname, "wt") as fp:
    json.dump(tests, fp, indent=4)



"""Tavakoli and Pezeshk (2005, :cite:`tavakoli05`) model."""

import numpy as np

from . import model

__author__ = "Albert Kottke"


class TavakoliPezeshk05(model.GroundMotionModel):
    """Tavakoli and Pezeshk (2005, :cite:`tavakoli05`) model.

    Developed for the Eastern North America with a reference velocity of 2880
    m/s.

    Parameters
    ----------
    scenario : :class:`pygmm.model.Scenario`
        earthquake scenario

    """

    NAME = "Tavakoli and Pezeshk (2005)"
    ABBREV = "TP05"

    # Reference velocity (m/sec)
    V_REF = 2880.0

    # Load the coefficients for the model
    COEFF = model.load_data_file("tavakoli_pezeshk_2005.csv", 1)
    PERIODS = COEFF["period"]

    INDEX_PGA = 0
    INDICES_PSA = np.arange(1, 14)

    PARAMS = [
        model.NumericParameter("dist_rup", True, None, 1000),
        model.NumericParameter("mag", True, 5.0, 8.2),
    ]

    def __init__(self, scenario: model.Scenario):
        """Initialize the model."""
        super().__init__(scenario)
        self._ln_resp = self._calc_ln_resp()
        self._ln_std = self._calc_ln_std()

    def _calc_ln_resp(self) -> np.ndarray:
        """Calculate the natural logarithm of the response.

        Returns
        -------
        ln_resp : class:`np.array`:
            natural log of the response

        """
        s = self._scenario
        c = self.COEFF

        # Magnitude scaling
        f1 = c.c_1 + c.c_2 * s.mag + c.c_3 * (8.5 - s.mag) ** 2.5

        # Distance scaling
        f2 = c.c_9 * np.log(s.dist_rup + 4.5)

        if s.dist_rup > 70:
            f2 += c.c_10 * np.log(s.dist_rup / 70.0)

        if s.dist_rup > 130:
            f2 += c.c_11 * np.log(s.dist_rup / 130.0)

        # Calculate scaled, magnitude dependent distance R for use when
        # calculating f3
        dist = np.sqrt(
            s.dist_rup**2 + (c.c_5 * np.exp(c.c_6 * s.mag + c.c_7 * (8.5 - s.mag) ** 2.5)) ** 2
        )
        f3 = (c.c_4 + c.c_13 * s.mag) * np.log(dist) + (c.c_8 + c.c_12 * s.mag) * dist

        # Compute the ground motion
        ln_resp = f1 + f2 + f3

        return ln_resp

    def _calc_ln_std(self) -> np.ndarray:
        """Calculate the logarithmic standard deviation.

        Returns
        -------
        ln_std : class:`np.array`:
            natural log standard deviation

        """
        s = self._scenario
        c = self.COEFF

        if s.mag < 7.2:
            ln_std = c.c_14 + c.c_15 * s.mag
        else:
            ln_std = c.c_16

        return ln_std



import os

import matplotlib.pyplot as plt

import pygmm

scenario = pygmm.Scenario(mag=6.5, dist_rup=30, v_s30=500, mechanism="SS", depth_tor=0)

models = [
    pygmm.BaylessAbrahamson2019(scenario),
    pygmm.BaylessAbrahamson2019(scenario.copy_with(v_s30=pygmm.BaylessAbrahamson2019.V_REF)),
]

fig, ax = plt.subplots()

for m in models:
    ax.plot(m.freqs, m.eas, label=f"{m.scenario.v_s30:.0f}")

ax.legend(title="$V_{s30}$ m/s")

ax.set(
    xlabel="Frequency (Hz)",
    xscale="log",
    ylabel="Effective Ampl. (g-s)",
    yscale="log",
)

fig.savefig(os.path.splitext(__file__)[0] + ".png", dpi=120)



"""Coppersmith and Bommer (2014) ground motion model, part of the PNNL Site
Wide Hazard."""

import numpy as np

from . import model
from .abrahamson_gregor_addo_2016 import AbrahamsonGregorAddo2016

__author__ = "Albert Kottke"


class CoppersmithBommer2014(AbrahamsonGregorAddo2016):
    """Coppersmith and Bommer (2014) ground motion model developed as part of
    the PNNL hazard study for the Hanford site.

    This model was developed for subduction regions and is a modified version
    of the BCHydro model.

    Parameters
    ----------
    scenario : :class:`pygmm.model.Scenario`
        earthquake scenario

    """

    NAME = "Coppersmith and Bommer (2014)"
    ABBREV = "CB14"

    # Load the coefficients for the model
    COEFF = model.load_data_file("coppersmith_bommer_2014.csv", 1)
    PERIODS = COEFF["period"]

    INDEX_PGA = 0
    INDEX_PGV = None
    INDICES_PSA = np.arange(14)

    # FIXME
    LIMITS = dict(
        mag=(3.0, 8.5),
        dist_jb=(0.0, 300.0),
        v_s30=(150.0, 1500.0),
    )

    PARAMS = [
        model.NumericParameter("mag", True, 3, 8.5),
        model.NumericParameter("dist_rup", False),
        model.NumericParameter("v_s30", True, 150.0, 1500.0),
        model.CategoricalParameter("event_type", True, ["interface", "intraslab"]),
    ]

    def __init__(self, scenario, scale_atten=1.0):
        """Initialize the model.

        Args:
            scenario (:class:`pygmm.model.Scenario`): earthquake scenario.
        """
        super().__init__(scenario, adjust_c1=None, adjust_c4=0, scale_atten=scale_atten)

    def _calc_ln_resp(self, pga_ref):
        """Calculate the natural logarithm of the response.

        Parameters
        ----------
        pga_ref : float
            peak ground acceleration (g) at the reference
            condition. If :class:`np.nan`, then no site term is applied.

        Returns
        -------
        ln_resp : class:`np.array`:
            natural log of the response

        """
        s = self._scenario
        c = self.COEFF

        path_atten = (c.t_2 + c.t_3 * (s.mag - 7.8)) * np.log(
            s.dist_rup + c.c_4 * np.exp(c.t_9 * (s.mag - 6))
        )

        ln_resp = (
            c.t_1
            + c.t_4 * self.adjust_c1
            + path_atten
            + self._scale_atten * c.t_6 * s.dist_rup
            + self._calc_f_mag(s.mag)
            + self._calc_f_site(pga_ref)
            + self._calc_f_faba(s.dist_rup)
        )

        return ln_resp

    def _calc_f_faba(self, dist):
        """Calculate the forearc/backarc scaling. Equation 4.

        Parameters
        ----------
        event_type : str
            type of subduction event, either: 'interface' or 'intraslab'

        dist : float
            depth to hypocenter [km]

        Returns
        -------
        f_faba : class:`np.array`:
            forearc/backarc scaling

        """
        c = self.COEFF
        f_faba = c.t_16 * np.log(np.maximum(dist, 40) / 40)

        return f_faba
